import sys

import pytest

from sentry_sdk.utils import event_from_exception

try:
    # Python 3.11
    from builtins import ExceptionGroup  # type: ignore
except ImportError:
    # Python 3.10 and below
    ExceptionGroup = None


minimum_python_311 = pytest.mark.skipif(
    sys.version_info < (3, 11), reason="ExceptionGroup tests need Python >= 3.11"
)


@minimum_python_311
def test_exceptiongroup():
    exception_group = None

    try:
        try:
            raise RuntimeError("something")
        except RuntimeError:
            raise ExceptionGroup(
                "nested",
                [
                    ValueError(654),
                    ExceptionGroup(
                        "imports",
                        [
                            ImportError("no_such_module"),
                            ModuleNotFoundError("another_module"),
                        ],
                    ),
                    TypeError("int"),
                ],
            )
    except ExceptionGroup as e:
        exception_group = e

    (event, _) = event_from_exception(
        exception_group,
        client_options={
            "include_local_variables": True,
            "include_source_context": True,
            "max_value_length": 1024,
        },
        mechanism={"type": "test_suite", "handled": False},
    )

    values = event["exception"]["values"]

    # For this test the stacktrace and the module is not important
    for x in values:
        if "stacktrace" in x:
            del x["stacktrace"]
        if "module" in x:
            del x["module"]

    expected_values = [
        {
            "mechanism": {
                "exception_id": 6,
                "handled": False,
                "parent_id": 0,
                "source": "exceptions[2]",
                "type": "chained",
            },
            "type": "TypeError",
            "value": "int",
        },
        {
            "mechanism": {
                "exception_id": 5,
                "handled": False,
                "parent_id": 3,
                "source": "exceptions[1]",
                "type": "chained",
            },
            "type": "ModuleNotFoundError",
            "value": "another_module",
        },
        {
            "mechanism": {
                "exception_id": 4,
                "handled": False,
                "parent_id": 3,
                "source": "exceptions[0]",
                "type": "chained",
            },
            "type": "ImportError",
            "value": "no_such_module",
        },
        {
            "mechanism": {
                "exception_id": 3,
                "handled": False,
                "is_exception_group": True,
                "parent_id": 0,
                "source": "exceptions[1]",
                "type": "chained",
            },
            "type": "ExceptionGroup",
            "value": "imports",
        },
        {
            "mechanism": {
                "exception_id": 2,
                "handled": False,
                "parent_id": 0,
                "source": "exceptions[0]",
                "type": "chained",
            },
            "type": "ValueError",
            "value": "654",
        },
        {
            "mechanism": {
                "exception_id": 1,
                "handled": False,
                "parent_id": 0,
                "source": "__context__",
                "type": "chained",
            },
            "type": "RuntimeError",
            "value": "something",
        },
        {
            "mechanism": {
                "exception_id": 0,
                "handled": False,
                "is_exception_group": True,
                "type": "test_suite",
            },
            "type": "ExceptionGroup",
            "value": "nested",
        },
    ]

    assert values == expected_values


@minimum_python_311
def test_exceptiongroup_simple():
    exception_group = None

    try:
        raise ExceptionGroup(
            "simple",
            [
                RuntimeError("something strange's going on"),
            ],
        )
    except ExceptionGroup as e:
        exception_group = e

    (event, _) = event_from_exception(
        exception_group,
        client_options={
            "include_local_variables": True,
            "include_source_context": True,
            "max_value_length": 1024,
        },
        mechanism={"type": "test_suite", "handled": False},
    )

    exception_values = event["exception"]["values"]

    assert len(exception_values) == 2

    assert exception_values[0]["type"] == "RuntimeError"
    assert exception_values[0]["value"] == "something strange's going on"
    assert exception_values[0]["mechanism"] == {
        "type": "chained",
        "handled": False,
        "exception_id": 1,
        "source": "exceptions[0]",
        "parent_id": 0,
    }

    assert exception_values[1]["type"] == "ExceptionGroup"
    assert exception_values[1]["value"] == "simple"
    assert exception_values[1]["mechanism"] == {
        "type": "test_suite",
        "handled": False,
        "exception_id": 0,
        "is_exception_group": True,
    }
    frame = exception_values[1]["stacktrace"]["frames"][0]
    assert frame["module"] == "tests.test_exceptiongroup"
    assert frame["context_line"] == "        raise ExceptionGroup("


@minimum_python_311
def test_exception_chain_cause():
    exception_chain_cause = ValueError("Exception with cause")
    exception_chain_cause.__context__ = TypeError("Exception in __context__")
    exception_chain_cause.__cause__ = TypeError(
        "Exception in __cause__"
    )  # this implicitly sets exception_chain_cause.__suppress_context__=True

    (event, _) = event_from_exception(
        exception_chain_cause,
        client_options={
            "include_local_variables": True,
            "include_source_context": True,
            "max_value_length": 1024,
        },
        mechanism={"type": "test_suite", "handled": False},
    )

    expected_exception_values = [
        {
            "mechanism": {
                "handled": False,
                "type": "test_suite",
            },
            "module": None,
            "type": "TypeError",
            "value": "Exception in __cause__",
        },
        {
            "mechanism": {
                "handled": False,
                "type": "test_suite",
            },
            "module": None,
            "type": "ValueError",
            "value": "Exception with cause",
        },
    ]

    exception_values = event["exception"]["values"]
    assert exception_values == expected_exception_values


@minimum_python_311
def test_exception_chain_context():
    exception_chain_context = ValueError("Exception with context")
    exception_chain_context.__context__ = TypeError("Exception in __context__")

    (event, _) = event_from_exception(
        exception_chain_context,
        client_options={
            "include_local_variables": True,
            "include_source_context": True,
            "max_value_length": 1024,
        },
        mechanism={"type": "test_suite", "handled": False},
    )

    expected_exception_values = [
        {
            "mechanism": {
                "handled": False,
                "type": "test_suite",
            },
            "module": None,
            "type": "TypeError",
            "value": "Exception in __context__",
        },
        {
            "mechanism": {
                "handled": False,
                "type": "test_suite",
            },
            "module": None,
            "type": "ValueError",
            "value": "Exception with context",
        },
    ]

    exception_values = event["exception"]["values"]
    assert exception_values == expected_exception_values


@minimum_python_311
def test_simple_exception():
    simple_excpetion = ValueError("A simple exception")

    (event, _) = event_from_exception(
        simple_excpetion,
        client_options={
            "include_local_variables": True,
            "include_source_context": True,
            "max_value_length": 1024,
        },
        mechanism={"type": "test_suite", "handled": False},
    )

    expected_exception_values = [
        {
            "mechanism": {
                "handled": False,
                "type": "test_suite",
            },
            "module": None,
            "type": "ValueError",
            "value": "A simple exception",
        },
    ]

    exception_values = event["exception"]["values"]
    assert exception_values == expected_exception_values


@minimum_python_311
def test_exceptiongroup_starlette_collapse():
    """
    Simulates the Starlette collapse_excgroups() pattern where a single-exception
    ExceptionGroup is caught and the inner exception is unwrapped and re-raised.

    See: https://github.com/Kludex/starlette/blob/0e88e92b592bfa11fd92e331869a8d49ba34b541/starlette/_utils.py#L79-L87

    When using FastAPI with multiple BaseHTTPMiddleware instances, anyio wraps
    exceptions in ExceptionGroups. Starlette's collapse_excgroups() then unwraps
    single-exception groups and re-raises the inner exception.

    When re-raising the unwrapped exception, Python implicitly sets __context__
    on it pointing back to the ExceptionGroup (because the re-raise happens
    inside the except block that caught the ExceptionGroup), creating a cycle:

        ExceptionGroup -> .exceptions[0] -> ValueError -> __context__ -> ExceptionGroup

    Without cycle detection in exceptions_from_error(), this causes infinite
    recursion and a silent RecursionError that drops the event.
    """
    exception_group = None

    try:
        try:
            raise RuntimeError("something")
        except RuntimeError:
            raise ExceptionGroup(
                "nested",
                [
                    ValueError(654),
                ],
            )
    except ExceptionGroup as exc:
        exception_group = exc
        unwrapped = exc.exceptions[0]
        try:
            raise unwrapped
        except Exception:
            pass

    (event, _) = event_from_exception(
        exception_group,
        client_options={
            "include_local_variables": True,
            "include_source_context": True,
            "max_value_length": 1024,
        },
        mechanism={"type": "test_suite", "handled": False},
    )

    values = event["exception"]["values"]

    # For this test the stacktrace and the module is not important
    for x in values:
        if "stacktrace" in x:
            del x["stacktrace"]
        if "module" in x:
            del x["module"]

    expected_values = [
        {
            "mechanism": {
                "exception_id": 2,
                "handled": False,
                "parent_id": 0,
                "source": "exceptions[0]",
                "type": "chained",
            },
            "type": "ValueError",
            "value": "654",
        },
        {
            "mechanism": {
                "exception_id": 1,
                "handled": False,
                "parent_id": 0,
                "source": "__context__",
                "type": "chained",
            },
            "type": "RuntimeError",
            "value": "something",
        },
        {
            "mechanism": {
                "exception_id": 0,
                "handled": False,
                "is_exception_group": True,
                "type": "test_suite",
            },
            "type": "ExceptionGroup",
            "value": "nested",
        },
    ]

    assert values == expected_values


@minimum_python_311
def test_cyclic_exception_group_cause():
    """
    Test case related to `test_exceptiongroup_starlette_collapse` above. We want to make sure that
    the same cyclic loop cannot happen via the __cause__ as well as the __context__
    """

    original = ValueError("original error")
    group = ExceptionGroup("unhandled errors in a TaskGroup", [original])
    original.__cause__ = group
    original.__suppress_context__ = True

    # When the ExceptionGroup is the top-level exception, exceptions_from_error
    # is called directly (not walk_exception_chain which has cycle detection).
    (event, _) = event_from_exception(
        group,
        client_options={
            "include_local_variables": True,
            "include_source_context": True,
            "max_value_length": 1024,
        },
        mechanism={"type": "test_suite", "handled": False},
    )

    exception_values = event["exception"]["values"]

    # Must produce a finite list of exceptions without hitting RecursionError.
    assert len(exception_values) >= 1
    exc_types = [v["type"] for v in exception_values]
    assert "ExceptionGroup" in exc_types
    assert "ValueError" in exc_types


@minimum_python_311
def test_deeply_nested_cyclic_exception_group():
    """
    Related to the `test_exceptiongroup_starlette_collapse` test above.

    Testing a more complex cycle: ExceptionGroup -> ValueError -> __cause__ ->
    ExceptionGroup (nested) -> TypeError -> __cause__ -> original ExceptionGroup
    """
    inner_error = TypeError("inner")
    outer_error = ValueError("outer")
    inner_group = ExceptionGroup("inner group", [inner_error])
    outer_group = ExceptionGroup("outer group", [outer_error])

    # Create a cycle spanning two ExceptionGroups
    outer_error.__cause__ = inner_group
    outer_error.__suppress_context__ = True
    inner_error.__cause__ = outer_group
    inner_error.__suppress_context__ = True

    (event, _) = event_from_exception(
        outer_group,
        client_options={
            "include_local_variables": True,
            "include_source_context": True,
            "max_value_length": 1024,
        },
        mechanism={"type": "test_suite", "handled": False},
    )

    exception_values = event["exception"]["values"]
    assert len(exception_values) >= 1
    exc_types = [v["type"] for v in exception_values]
    assert "ExceptionGroup" in exc_types
    assert "ValueError" in exc_types
    assert "TypeError" in exc_types



import re
from typing import TYPE_CHECKING
from unittest import mock

import pytest
import responses
from huggingface_hub import InferenceClient

import sentry_sdk
from sentry_sdk.integrations.huggingface_hub import HuggingfaceHubIntegration
from sentry_sdk.utils import package_version, safe_serialize

try:
    from huggingface_hub.utils._errors import HfHubHTTPError
except ImportError:
    from huggingface_hub.errors import HfHubHTTPError


if TYPE_CHECKING:
    from typing import Any


HF_VERSION = package_version("huggingface-hub")

if HF_VERSION and HF_VERSION < (0, 30, 0):
    MODEL_ENDPOINT = "https://api-inference.huggingface.co/models/{model_name}"
    INFERENCE_ENDPOINT = "https://api-inference.huggingface.co/models/{model_name}"
else:
    MODEL_ENDPOINT = "https://huggingface.co/api/models/{model_name}"
    INFERENCE_ENDPOINT = (
        "https://router.huggingface.co/hf-inference/models/{model_name}"
    )


def get_hf_provider_inference_client():
    # The provider parameter was added in version 0.28.0 of huggingface_hub
    return (
        InferenceClient(model="test-model", provider="hf-inference")
        if HF_VERSION >= (0, 28, 0)
        else InferenceClient(model="test-model")
    )


def _add_mock_response(
    httpx_mock, rsps, method, url, json=None, status=200, body=None, headers=None
):
    # HF v1+ uses httpx for making requests to their API, while <1 uses requests.
    # Since we have to test both, we need mocks for both httpx and requests.
    if HF_VERSION >= (1, 0, 0):
        httpx_mock.add_response(
            method=method,
            url=url,
            json=json,
            content=body,
            status_code=status,
            headers=headers,
            is_optional=True,
            is_reusable=True,
        )
    else:
        rsps.add(
            method=method,
            url=url,
            json=json,
            body=body,
            status=status,
            headers=headers,
        )


@pytest.fixture
def mock_hf_text_generation_api(httpx_mock):
    # type: () -> Any
    """Mock HuggingFace text generation API"""

    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        model_name = "test-model"

        _add_mock_response(
            httpx_mock,
            rsps,
            "GET",
            re.compile(
                MODEL_ENDPOINT.format(model_name=model_name)
                + r"(\?expand=inferenceProviderMapping)?"
            ),
            json={
                "id": model_name,
                "pipeline_tag": "text-generation",
                "inferenceProviderMapping": {
                    "hf-inference": {
                        "status": "live",
                        "providerId": model_name,
                        "task": "text-generation",
                    }
                },
            },
            status=200,
        )

        _add_mock_response(
            httpx_mock,
            rsps,
            "POST",
            INFERENCE_ENDPOINT.format(model_name=model_name),
            json={
                "generated_text": "[mocked] Hello! How can i help you?",
                "details": {
                    "finish_reason": "length",
                    "generated_tokens": 10,
                    "prefill": [],
                    "tokens": [],
                },
            },
            status=200,
        )

        if HF_VERSION >= (1, 0, 0):
            yield httpx_mock
        else:
            yield rsps


@pytest.fixture
def mock_hf_api_with_errors(httpx_mock):
    # type: () -> Any
    """Mock HuggingFace API that always raises errors for any request"""

    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        model_name = "test-model"

        # Mock model info endpoint with error
        _add_mock_response(
            httpx_mock,
            rsps,
            "GET",
            MODEL_ENDPOINT.format(model_name=model_name),
            json={"error": "Model not found"},
            status=404,
        )

        # Mock text generation endpoint with error
        _add_mock_response(
            httpx_mock,
            rsps,
            "POST",
            INFERENCE_ENDPOINT.format(model_name=model_name),
            json={"error": "Internal server error", "message": "Something went wrong"},
            status=500,
        )

        # Mock chat completion endpoint with error
        _add_mock_response(
            httpx_mock,
            rsps,
            "POST",
            INFERENCE_ENDPOINT.format(model_name=model_name) + "/v1/chat/completions",
            json={"error": "Internal server error", "message": "Something went wrong"},
            status=500,
        )

        # Catch-all pattern for any other model requests
        _add_mock_response(
            httpx_mock,
            rsps,
            "GET",
            "https://huggingface.co/api/models/test-model-error",
            json={"error": "Generic model error"},
            status=500,
        )

        if HF_VERSION >= (1, 0, 0):
            yield httpx_mock
        else:
            yield rsps


@pytest.fixture
def mock_hf_text_generation_api_streaming(httpx_mock):
    # type: () -> Any
    """Mock streaming HuggingFace text generation API"""
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        model_name = "test-model"

        # Mock model info endpoint
        _add_mock_response(
            httpx_mock,
            rsps,
            "GET",
            MODEL_ENDPOINT.format(model_name=model_name),
            json={
                "id": model_name,
                "pipeline_tag": "text-generation",
                "inferenceProviderMapping": {
                    "hf-inference": {
                        "status": "live",
                        "providerId": model_name,
                        "task": "text-generation",
                    }
                },
            },
            status=200,
        )

        # Mock text generation endpoint for streaming
        streaming_response = b'data:{"token":{"id":1, "special": false, "text": "the mocked "}}\n\ndata:{"token":{"id":2, "special": false, "text": "model response"}, "details":{"finish_reason": "length", "generated_tokens": 10, "seed": 0}}\n\n'

        _add_mock_response(
            httpx_mock,
            rsps,
            "POST",
            INFERENCE_ENDPOINT.format(model_name=model_name),
            body=streaming_response,
            status=200,
            headers={
                "Content-Type": "text/event-stream",
                "Cache-Control": "no-cache",
                "Connection": "keep-alive",
            },
        )

        if HF_VERSION >= (1, 0, 0):
            yield httpx_mock
        else:
            yield rsps


@pytest.fixture
def mock_hf_chat_completion_api(httpx_mock):
    # type: () -> Any
    """Mock HuggingFace chat completion API"""
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        model_name = "test-model"

        # Mock model info endpoint
        _add_mock_response(
            httpx_mock,
            rsps,
            "GET",
            MODEL_ENDPOINT.format(model_name=model_name),
            json={
                "id": model_name,
                "pipeline_tag": "conversational",
                "inferenceProviderMapping": {
                    "hf-inference": {
                        "status": "live",
                        "providerId": model_name,
                        "task": "conversational",
                    }
                },
            },
            status=200,
        )

        # Mock chat completion endpoint
        _add_mock_response(
            httpx_mock,
            rsps,
            "POST",
            INFERENCE_ENDPOINT.format(model_name=model_name) + "/v1/chat/completions",
            json={
                "id": "xyz-123",
                "created": 1234567890,
                "model": f"{model_name}-123",
                "system_fingerprint": "fp_123",
                "choices": [
                    {
                        "index": 0,
                        "finish_reason": "stop",
                        "message": {
                            "role": "assistant",
                            "content": "[mocked] Hello! How can I help you today?",
                        },
                    }
                ],
                "usage": {
                    "completion_tokens": 8,
                    "prompt_tokens": 10,
                    "total_tokens": 18,
                },
            },
            status=200,
        )

        if HF_VERSION >= (1, 0, 0):
            yield httpx_mock
        else:
            yield rsps


@pytest.fixture
def mock_hf_chat_completion_api_tools(httpx_mock):
    # type: () -> Any
    """Mock HuggingFace chat completion API with tool calls."""
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        model_name = "test-model"

        # Mock model info endpoint
        _add_mock_response(
            httpx_mock,
            rsps,
            "GET",
            MODEL_ENDPOINT.format(model_name=model_name),
            json={
                "id": model_name,
                "pipeline_tag": "conversational",
                "inferenceProviderMapping": {
                    "hf-inference": {
                        "status": "live",
                        "providerId": model_name,
                        "task": "conversational",
                    }
                },
            },
            status=200,
        )

        # Mock chat completion endpoint
        _add_mock_response(
            httpx_mock,
            rsps,
            "POST",
            INFERENCE_ENDPOINT.format(model_name=model_name) + "/v1/chat/completions",
            json={
                "id": "xyz-123",
                "created": 1234567890,
                "model": f"{model_name}-123",
                "system_fingerprint": "fp_123",
                "choices": [
                    {
                        "index": 0,
                        "finish_reason": "tool_calls",
                        "message": {
                            "role": "assistant",
                            "tool_calls": [
                                {
                                    "id": "call_123",
                                    "type": "function",
                                    "function": {
                                        "name": "get_weather",
                                        "arguments": {"location": "Paris"},
                                    },
                                }
                            ],
                        },
                    }
                ],
                "usage": {
                    "completion_tokens": 8,
                    "prompt_tokens": 10,
                    "total_tokens": 18,
                },
            },
            status=200,
        )

        if HF_VERSION >= (1, 0, 0):
            yield httpx_mock
        else:
            yield rsps


@pytest.fixture
def mock_hf_chat_completion_api_streaming(httpx_mock):
    # type: () -> Any
    """Mock streaming HuggingFace chat completion API"""
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        model_name = "test-model"

        # Mock model info endpoint
        _add_mock_response(
            httpx_mock,
            rsps,
            "GET",
            MODEL_ENDPOINT.format(model_name=model_name),
            json={
                "id": model_name,
                "pipeline_tag": "conversational",
                "inferenceProviderMapping": {
                    "hf-inference": {
                        "status": "live",
                        "providerId": model_name,
                        "task": "conversational",
                    }
                },
            },
            status=200,
        )

        # Mock chat completion streaming endpoint
        streaming_chat_response = (
            b'data:{"id":"xyz-123","created":1234567890,"model":"test-model-123","system_fingerprint":"fp_123","choices":[{"delta":{"role":"assistant","content":"the mocked "},"index":0,"finish_reason":null}],"usage":null}\n\n'
            b'data:{"id":"xyz-124","created":1234567890,"model":"test-model-123","system_fingerprint":"fp_123","choices":[{"delta":{"role":"assistant","content":"model response"},"index":0,"finish_reason":"stop"}],"usage":{"prompt_tokens":183,"completion_tokens":14,"total_tokens":197}}\n\n'
        )

        _add_mock_response(
            httpx_mock,
            rsps,
            "POST",
            INFERENCE_ENDPOINT.format(model_name=model_name) + "/v1/chat/completions",
            body=streaming_chat_response,
            status=200,
            headers={
                "Content-Type": "text/event-stream",
                "Cache-Control": "no-cache",
                "Connection": "keep-alive",
            },
        )

        if HF_VERSION >= (1, 0, 0):
            yield httpx_mock
        else:
            yield rsps


@pytest.fixture
def mock_hf_chat_completion_api_streaming_tools(httpx_mock):
    # type: () -> Any
    """Mock streaming HuggingFace chat completion API with tool calls."""
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        model_name = "test-model"

        # Mock model info endpoint
        _add_mock_response(
            httpx_mock,
            rsps,
            "GET",
            MODEL_ENDPOINT.format(model_name=model_name),
            json={
                "id": model_name,
                "pipeline_tag": "conversational",
                "inferenceProviderMapping": {
                    "hf-inference": {
                        "status": "live",
                        "providerId": model_name,
                        "task": "conversational",
                    }
                },
            },
            status=200,
        )

        # Mock chat completion streaming endpoint
        streaming_chat_response = (
            b'data:{"id":"xyz-123","created":1234567890,"model":"test-model-123","system_fingerprint":"fp_123","choices":[{"delta":{"role":"assistant","content":"response with tool calls follows"},"index":0,"finish_reason":null}],"usage":null}\n\n'
            b'data:{"id":"xyz-124","created":1234567890,"model":"test-model-123","system_fingerprint":"fp_123","choices":[{"delta":{"role":"assistant","tool_calls": [{"id": "call_123","type": "function","function": {"name": "get_weather", "arguments": {"location": "Paris"}}}]},"index":0,"finish_reason":"tool_calls"}],"usage":{"prompt_tokens":183,"completion_tokens":14,"total_tokens":197}}\n\n'
        )

        _add_mock_response(
            httpx_mock,
            rsps,
            "POST",
            INFERENCE_ENDPOINT.format(model_name=model_name) + "/v1/chat/completions",
            body=streaming_chat_response,
            status=200,
            headers={
                "Content-Type": "text/event-stream",
                "Cache-Control": "no-cache",
                "Connection": "keep-alive",
            },
        )

        if HF_VERSION >= (1, 0, 0):
            yield httpx_mock
        else:
            yield rsps


@pytest.mark.parametrize("stream_gen_ai_spans", [True, False])
@pytest.mark.httpx_mock(assert_all_requests_were_expected=False)
@pytest.mark.parametrize("send_default_pii", [True, False])
@pytest.mark.parametrize("include_prompts", [True, False])
def test_text_generation(
    sentry_init: "Any",
    capture_events: "Any",
    capture_items: "Any",
    send_default_pii: "Any",
    include_prompts: "Any",
    mock_hf_text_generation_api: "Any",
    stream_gen_ai_spans: "Any",
) -> None:
    sentry_init(
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        integrations=[HuggingfaceHubIntegration(include_prompts=include_prompts)],
        stream_gen_ai_spans=stream_gen_ai_spans,
    )

    client = InferenceClient(model="test-model")

    if stream_gen_ai_spans:
        items = capture_items("transaction", "span")

        with sentry_sdk.start_transaction(name="test"):
            client.text_generation(
                "Hello",
                stream=False,
                details=True,
            )

        spans = [item.payload for item in items if item.type == "span"]
        span = None
        for sp in spans:
            if sp["attributes"]["sentry.op"].startswith("gen_ai"):
                assert span is None, "there is exactly one gen_ai span"
                span = sp
            else:
                # there should be no other spans, just the gen_ai span
                # and optionally some http.client spans from talking to the hf api
                assert sp["attributes"]["sentry.op"] == "http.client"

        assert span is not None

        assert span["attributes"]["sentry.op"] == "gen_ai.text_completion"
        assert span["name"] == "text_completion test-model"
        assert span["attributes"]["sentry.origin"] == "auto.ai.huggingface_hub"

        expected_data = {
            "gen_ai.operation.name": "text_completion",
            "gen_ai.request.model": "test-model",
            "gen_ai.response.finish_reasons": "length",
            "gen_ai.response.streaming": False,
            "gen_ai.usage.total_tokens": 10,
            "process.runtime.name": mock.ANY,
            "process.runtime.version": mock.ANY,
            "sentry.environment": "production",
            "sentry.op": "gen_ai.text_completion",
            "sentry.origin": "auto.ai.huggingface_hub",
            "sentry.release": mock.ANY,
            "sentry.sdk.name": "sentry.python",
            "sentry.sdk.version": mock.ANY,
            "sentry.segment.id": mock.ANY,
            "sentry.segment.name": "test",
            "server.address": mock.ANY,
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }

        if send_default_pii and include_prompts:
            expected_data["gen_ai.request.messages"] = "Hello"
            expected_data["gen_ai.response.text"] = (
                "[mocked] Hello! How can i help you?"
            )

        if not send_default_pii or not include_prompts:
            assert "gen_ai.request.messages" not in expected_data
            assert "gen_ai.response.text" not in expected_data

        assert span["attributes"] == expected_data

        # text generation does not set the response model
        assert "gen_ai.response.model" not in span["attributes"]
    else:
        events = capture_events()

        with sentry_sdk.start_transaction(name="test"):
            client.text_generation(
                "Hello",
                stream=False,
                details=True,
            )

        (transaction,) = events

        span = None
        for sp in transaction["spans"]:
            if sp["op"].startswith("gen_ai"):
                assert span is None, "there is exactly one gen_ai span"
                span = sp
            else:
                # there should be no other spans, just the gen_ai span
                # and optionally some http.client spans from talking to the hf api
                assert sp["op"] == "http.client"

        assert span is not None

        assert span["op"] == "gen_ai.text_completion"
        assert span["description"] == "text_completion test-model"
        assert span["origin"] == "auto.ai.huggingface_hub"

        expected_data = {
            "gen_ai.operation.name": "text_completion",
            "gen_ai.request.model": "test-model",
            "gen_ai.response.finish_reasons": "length",
            "gen_ai.response.streaming": False,
            "gen_ai.usage.total_tokens": 10,
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }

        if send_default_pii and include_prompts:
            expected_data["gen_ai.request.messages"] = "Hello"
            expected_data["gen_ai.response.text"] = (
                "[mocked] Hello! How can i help you?"
            )

        if not send_default_pii or not include_prompts:
            assert "gen_ai.request.messages" not in expected_data
            assert "gen_ai.response.text" not in expected_data

        assert span["data"] == expected_data

        # text generation does not set the response model
        assert "gen_ai.response.model" not in span["data"]


@pytest.mark.parametrize("stream_gen_ai_spans", [True, False])
@pytest.mark.httpx_mock(assert_all_requests_were_expected=False)
@pytest.mark.parametrize("send_default_pii", [True, False])
@pytest.mark.parametrize("include_prompts", [True, False])
def test_text_generation_streaming(
    sentry_init: "Any",
    capture_events: "Any",
    capture_items: "Any",
    send_default_pii: "Any",
    include_prompts: "Any",
    mock_hf_text_generation_api_streaming: "Any",
    stream_gen_ai_spans: "Any",
) -> None:
    sentry_init(
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        integrations=[HuggingfaceHubIntegration(include_prompts=include_prompts)],
        stream_gen_ai_spans=stream_gen_ai_spans,
    )

    client = InferenceClient(model="test-model")

    if stream_gen_ai_spans:
        items = capture_items("transaction", "span")

        with sentry_sdk.start_transaction(name="test"):
            for _ in client.text_generation(
                prompt="Hello",
                stream=True,
                details=True,
            ):
                pass

        spans = [item.payload for item in items if item.type == "span"]
        span = None
        for sp in spans:
            if sp["attributes"]["sentry.op"].startswith("gen_ai"):
                assert span is None, "there is exactly one gen_ai span"
                span = sp
            else:
                # there should be no other spans, just the gen_ai span
                # and optionally some http.client spans from talking to the hf api
                assert sp["attributes"]["sentry.op"] == "http.client"

        assert span is not None

        assert span["attributes"]["sentry.op"] == "gen_ai.text_completion"
        assert span["name"] == "text_completion test-model"
        assert span["attributes"]["sentry.origin"] == "auto.ai.huggingface_hub"

        expected_data = {
            "gen_ai.operation.name": "text_completion",
            "gen_ai.request.model": "test-model",
            "gen_ai.response.finish_reasons": "length",
            "gen_ai.response.streaming": True,
            "gen_ai.usage.total_tokens": 10,
            "process.runtime.name": mock.ANY,
            "process.runtime.version": mock.ANY,
            "sentry.environment": "production",
            "sentry.op": "gen_ai.text_completion",
            "sentry.origin": "auto.ai.huggingface_hub",
            "sentry.release": mock.ANY,
            "sentry.sdk.name": "sentry.python",
            "sentry.sdk.version": mock.ANY,
            "sentry.segment.id": mock.ANY,
            "sentry.segment.name": "test",
            "server.address": mock.ANY,
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }

        if send_default_pii and include_prompts:
            expected_data["gen_ai.request.messages"] = "Hello"
            expected_data["gen_ai.response.text"] = "the mocked model response"

        if not send_default_pii or not include_prompts:
            assert "gen_ai.request.messages" not in expected_data
            assert "gen_ai.response.text" not in expected_data

        assert span["attributes"] == expected_data

        # text generation does not set the response model
        assert "gen_ai.response.model" not in span["attributes"]
    else:
        events = capture_events()

        with sentry_sdk.start_transaction(name="test"):
            for _ in client.text_generation(
                prompt="Hello",
                stream=True,
                details=True,
            ):
                pass

        (transaction,) = events

        span = None
        for sp in transaction["spans"]:
            if sp["op"].startswith("gen_ai"):
                assert span is None, "there is exactly one gen_ai span"
                span = sp
            else:
                # there should be no other spans, just the gen_ai span
                # and optionally some http.client spans from talking to the hf api
                assert sp["op"] == "http.client"

        assert span is not None

        assert span["op"] == "gen_ai.text_completion"
        assert span["description"] == "text_completion test-model"
        assert span["origin"] == "auto.ai.huggingface_hub"

        expected_data = {
            "gen_ai.operation.name": "text_completion",
            "gen_ai.request.model": "test-model",
            "gen_ai.response.finish_reasons": "length",
            "gen_ai.response.streaming": True,
            "gen_ai.usage.total_tokens": 10,
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }

        if send_default_pii and include_prompts:
            expected_data["gen_ai.request.messages"] = "Hello"
            expected_data["gen_ai.response.text"] = "the mocked model response"

        if not send_default_pii or not include_prompts:
            assert "gen_ai.request.messages" not in expected_data
            assert "gen_ai.response.text" not in expected_data

        assert span["data"] == expected_data

        # text generation does not set the response model
        assert "gen_ai.response.model" not in span["data"]


@pytest.mark.parametrize("stream_gen_ai_spans", [True, False])
@pytest.mark.httpx_mock(assert_all_requests_were_expected=False)
@pytest.mark.parametrize("send_default_pii", [True, False])
@pytest.mark.parametrize("include_prompts", [True, False])
def test_chat_completion(
    sentry_init: "Any",
    capture_events: "Any",
    capture_items: "Any",
    send_default_pii: "Any",
    include_prompts: "Any",
    mock_hf_chat_completion_api: "Any",
    stream_gen_ai_spans: "Any",
) -> None:
    sentry_init(
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        integrations=[HuggingfaceHubIntegration(include_prompts=include_prompts)],
        stream_gen_ai_spans=stream_gen_ai_spans,
    )

    client = get_hf_provider_inference_client()

    if stream_gen_ai_spans:
        items = capture_items("transaction", "span")

        with sentry_sdk.start_transaction(name="test"):
            client.chat_completion(
                messages=[
                    {
                        "role": "user",
                        "content": "Message demonstrating the absence of truncation.",
                    },
                    {"role": "user", "content": "Hello!"},
                ],
                stream=False,
            )

        spans = [item.payload for item in items if item.type == "span"]
        span = None
        for sp in spans:
            if sp["attributes"]["sentry.op"].startswith("gen_ai"):
                assert span is None, "there is exactly one gen_ai span"
                span = sp
            else:
                # there should be no other spans, just the gen_ai span
                # and optionally some http.client spans from talking to the hf api
                assert sp["attributes"]["sentry.op"] == "http.client"

        assert span is not None

        assert span["attributes"]["sentry.op"] == "gen_ai.chat"
        assert span["name"] == "chat test-model"
        assert span["attributes"]["sentry.origin"] == "auto.ai.huggingface_hub"

        expected_data = {
            "gen_ai.operation.name": "chat",
            "gen_ai.request.model": "test-model",
            "gen_ai.response.finish_reasons": "stop",
            "gen_ai.response.model": "test-model-123",
            "gen_ai.response.streaming": False,
            "gen_ai.usage.input_tokens": 10,
            "gen_ai.usage.output_tokens": 8,
            "gen_ai.usage.total_tokens": 18,
            "process.runtime.name": mock.ANY,
            "process.runtime.version": mock.ANY,
            "sentry.environment": "production",
            "sentry.op": "gen_ai.chat",
            "sentry.origin": "auto.ai.huggingface_hub",
            "sentry.release": mock.ANY,
            "sentry.sdk.name": "sentry.python",
            "sentry.sdk.version": mock.ANY,
            "sentry.segment.id": mock.ANY,
            "sentry.segment.name": "test",
            "server.address": mock.ANY,
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }

        if send_default_pii and include_prompts:
            expected_data["gen_ai.request.messages"] = safe_serialize(
                [
                    {
                        "role": "user",
                        "content": "Message demonstrating the absence of truncation.",
                    },
                    {"role": "user", "content": "Hello!"},
                ]
            )
            expected_data["gen_ai.response.text"] = (
                "[mocked] Hello! How can I help you today?"
            )

        if not send_default_pii or not include_prompts:
            assert "gen_ai.request.messages" not in expected_data
            assert "gen_ai.response.text" not in expected_data

        assert span["attributes"] == expected_data
    else:
        events = capture_events()

        with sentry_sdk.start_transaction(name="test"):
            client.chat_completion(
                messages=[{"role": "user", "content": "Hello!"}],
                stream=False,
            )

        (transaction,) = events

        span = None
        for sp in transaction["spans"]:
            if sp["op"].startswith("gen_ai"):
                assert span is None, "there is exactly one gen_ai span"
                span = sp
            else:
                # there should be no other spans, just the gen_ai span
                # and optionally some http.client spans from talking to the hf api
                assert sp["op"] == "http.client"

        assert span is not None

        assert span["op"] == "gen_ai.chat"
        assert span["description"] == "chat test-model"
        assert span["origin"] == "auto.ai.huggingface_hub"

        expected_data = {
            "gen_ai.operation.name": "chat",
            "gen_ai.request.model": "test-model",
            "gen_ai.response.finish_reasons": "stop",
            "gen_ai.response.model": "test-model-123",
            "gen_ai.response.streaming": False,
            "gen_ai.usage.input_tokens": 10,
            "gen_ai.usage.output_tokens": 8,
            "gen_ai.usage.total_tokens": 18,
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }

        if send_default_pii and include_prompts:
            expected_data["gen_ai.request.messages"] = (
                '[{"role": "user", "content": "Hello!"}]'
            )
            expected_data["gen_ai.response.text"] = (
                "[mocked] Hello! How can I help you today?"
            )

        if not send_default_pii or not include_prompts:
            assert "gen_ai.request.messages" not in expected_data
            assert "gen_ai.response.text" not in expected_data

        assert span["data"] == expected_data


@pytest.mark.parametrize("stream_gen_ai_spans", [True, False])
@pytest.mark.httpx_mock(assert_all_requests_were_expected=False)
@pytest.mark.parametrize("send_default_pii", [True, False])
@pytest.mark.parametrize("include_prompts", [True, False])
def test_chat_completion_streaming(
    sentry_init: "Any",
    capture_events: "Any",
    capture_items: "Any",
    send_default_pii: "Any",
    include_prompts: "Any",
    mock_hf_chat_completion_api_streaming: "Any",
    stream_gen_ai_spans: "Any",
) -> None:
    sentry_init(
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        integrations=[HuggingfaceHubIntegration(include_prompts=include_prompts)],
        stream_gen_ai_spans=stream_gen_ai_spans,
    )

    client = get_hf_provider_inference_client()

    if stream_gen_ai_spans:
        items = capture_items("transaction", "span")

        with sentry_sdk.start_transaction(name="test"):
            _ = list(
                client.chat_completion(
                    [
                        {
                            "role": "user",
                            "content": "Message demonstrating the absence of truncation.",
                        },
                        {"role": "user", "content": "Hello!"},
                    ],
                    stream=True,
                )
            )

        spans = [item.payload for item in items if item.type == "span"]
        span = None
        for sp in spans:
            if sp["attributes"]["sentry.op"].startswith("gen_ai"):
                assert span is None, "there is exactly one gen_ai span"
                span = sp
            else:
                # there should be no other spans, just the gen_ai span
                # and optionally some http.client spans from talking to the hf api
                assert sp["attributes"]["sentry.op"] == "http.client"

        assert span is not None

        assert span["attributes"]["sentry.op"] == "gen_ai.chat"
        assert span["name"] == "chat test-model"
        assert span["attributes"]["sentry.origin"] == "auto.ai.huggingface_hub"

        expected_data = {
            "gen_ai.operation.name": "chat",
            "gen_ai.request.model": "test-model",
            "gen_ai.response.finish_reasons": "stop",
            "gen_ai.response.model": "test-model-123",
            "gen_ai.response.streaming": True,
            "process.runtime.name": mock.ANY,
            "process.runtime.version": mock.ANY,
            "sentry.environment": "production",
            "sentry.op": "gen_ai.chat",
            "sentry.origin": "auto.ai.huggingface_hub",
            "sentry.release": mock.ANY,
            "sentry.sdk.name": "sentry.python",
            "sentry.sdk.version": mock.ANY,
            "sentry.segment.id": mock.ANY,
            "sentry.segment.name": "test",
            "server.address": mock.ANY,
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }
        # usage is not available in older versions of the library
        if HF_VERSION and HF_VERSION >= (0, 26, 0):
            expected_data["gen_ai.usage.input_tokens"] = 183
            expected_data["gen_ai.usage.output_tokens"] = 14
            expected_data["gen_ai.usage.total_tokens"] = 197

        if send_default_pii and include_prompts:
            expected_data["gen_ai.request.messages"] = safe_serialize(
                [
                    {
                        "role": "user",
                        "content": "Message demonstrating the absence of truncation.",
                    },
                    {"role": "user", "content": "Hello!"},
                ]
            )
            expected_data["gen_ai.response.text"] = "the mocked model response"

        if not send_default_pii or not include_prompts:
            assert "gen_ai.request.messages" not in expected_data
            assert "gen_ai.response.text" not in expected_data

        assert span["attributes"] == expected_data
    else:
        events = capture_events()

        with sentry_sdk.start_transaction(name="test"):
            _ = list(
                client.chat_completion(
                    [{"role": "user", "content": "Hello!"}],
                    stream=True,
                )
            )

        (transaction,) = events

        span = None
        for sp in transaction["spans"]:
            if sp["op"].startswith("gen_ai"):
                assert span is None, "there is exactly one gen_ai span"
                span = sp
            else:
                # there should be no other spans, just the gen_ai span
                # and optionally some http.client spans from talking to the hf api
                assert sp["op"] == "http.client"

        assert span is not None

        assert span["op"] == "gen_ai.chat"
        assert span["description"] == "chat test-model"
        assert span["origin"] == "auto.ai.huggingface_hub"

        expected_data = {
            "gen_ai.operation.name": "chat",
            "gen_ai.request.model": "test-model",
            "gen_ai.response.finish_reasons": "stop",
            "gen_ai.response.model": "test-model-123",
            "gen_ai.response.streaming": True,
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }
        # usage is not available in older versions of the library
        if HF_VERSION and HF_VERSION >= (0, 26, 0):
            expected_data["gen_ai.usage.input_tokens"] = 183
            expected_data["gen_ai.usage.output_tokens"] = 14
            expected_data["gen_ai.usage.total_tokens"] = 197

        if send_default_pii and include_prompts:
            expected_data["gen_ai.request.messages"] = (
                '[{"role": "user", "content": "Hello!"}]'
            )
            expected_data["gen_ai.response.text"] = "the mocked model response"

        if not send_default_pii or not include_prompts:
            assert "gen_ai.request.messages" not in expected_data
            assert "gen_ai.response.text" not in expected_data

        assert span["data"] == expected_data


@pytest.mark.parametrize("stream_gen_ai_spans", [True, False])
@pytest.mark.httpx_mock(assert_all_requests_were_expected=False)
def test_chat_completion_api_error(
    sentry_init: "Any",
    capture_events: "Any",
    capture_items: "Any",
    mock_hf_api_with_errors: "Any",
    stream_gen_ai_spans: "Any",
) -> None:
    sentry_init(
        traces_sample_rate=1.0,
        stream_gen_ai_spans=stream_gen_ai_spans,
    )

    client = get_hf_provider_inference_client()
    if stream_gen_ai_spans:
        items = capture_items("event", "transaction", "span")

        with sentry_sdk.start_transaction(name="test"), pytest.raises(HfHubHTTPError):
            client.chat_completion(
                messages=[{"role": "user", "content": "Hello!"}],
            )

        (error,) = (item.payload for item in items if item.type == "event")

        assert error["exception"]["values"][0]["mechanism"]["type"] == "huggingface_hub"
        assert not error["exception"]["values"][0]["mechanism"]["handled"]

        spans = [item.payload for item in items if item.type == "span"]
        span = None
        for sp in spans:
            if sp["attributes"]["sentry.op"].startswith("gen_ai"):
                assert span is None, "there is exactly one gen_ai span"
                span = sp
            else:
                # there should be no other spans, just the gen_ai span
                # and optionally some http.client spans from talking to the hf api
                assert sp["attributes"]["sentry.op"] == "http.client"

        assert span is not None

        assert span["attributes"]["sentry.op"] == "gen_ai.chat"
        assert span["name"] == "chat test-model"
        assert span["attributes"]["sentry.origin"] == "auto.ai.huggingface_hub"
        assert span["status"] == "error"

        (transaction,) = (item.payload for item in items if item.type == "transaction")

        assert (
            error["contexts"]["trace"]["trace_id"]
            == transaction["contexts"]["trace"]["trace_id"]
        )
        expected_data = {
            "gen_ai.operation.name": "chat",
            "gen_ai.request.model": "test-model",
            "process.runtime.name": mock.ANY,
            "process.runtime.version": mock.ANY,
            "sentry.environment": "production",
            "sentry.op": "gen_ai.chat",
            "sentry.origin": "auto.ai.huggingface_hub",
            "sentry.release": mock.ANY,
            "sentry.sdk.name": "sentry.python",
            "sentry.sdk.version": mock.ANY,
            "sentry.segment.id": mock.ANY,
            "sentry.segment.name": "test",
            "server.address": mock.ANY,
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }
        assert span["attributes"] == expected_data
    else:
        events = capture_events()

        with sentry_sdk.start_transaction(name="test"), pytest.raises(HfHubHTTPError):
            client.chat_completion(
                messages=[{"role": "user", "content": "Hello!"}],
            )

        (
            error,
            transaction,
        ) = events

        assert error["exception"]["values"][0]["mechanism"]["type"] == "huggingface_hub"
        assert not error["exception"]["values"][0]["mechanism"]["handled"]

        span = None
        for sp in transaction["spans"]:
            if sp["op"].startswith("gen_ai"):
                assert span is None, "there is exactly one gen_ai span"
                span = sp
            else:
                # there should be no other spans, just the gen_ai span
                # and optionally some http.client spans from talking to the hf api
                assert sp["op"] == "http.client"

        assert span is not None

        assert span["op"] == "gen_ai.chat"
        assert span["description"] == "chat test-model"
        assert span["origin"] == "auto.ai.huggingface_hub"
        assert span["status"] == "internal_error"
        assert span.get("tags", {}).get("status") == "internal_error"

        assert (
            error["contexts"]["trace"]["trace_id"]
            == transaction["contexts"]["trace"]["trace_id"]
        )
        expected_data = {
            "gen_ai.operation.name": "chat",
            "gen_ai.request.model": "test-model",
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }
        assert span["data"] == expected_data


@pytest.mark.parametrize("stream_gen_ai_spans", [True, False])
@pytest.mark.httpx_mock(assert_all_requests_were_expected=False)
def test_span_status_error(
    sentry_init: "Any",
    capture_events: "Any",
    capture_items: "Any",
    mock_hf_api_with_errors: "Any",
    stream_gen_ai_spans: "Any",
) -> None:
    client = get_hf_provider_inference_client()

    sentry_init(
        traces_sample_rate=1.0,
        stream_gen_ai_spans=stream_gen_ai_spans,
    )

    if stream_gen_ai_spans:
        items = capture_items("event", "transaction", "span")

        with sentry_sdk.start_transaction(name="test"), pytest.raises(HfHubHTTPError):
            client.chat_completion(
                messages=[{"role": "user", "content": "Hello!"}],
            )

        (error,) = [item.payload for item in items if item.type == "event"]
        assert error["level"] == "error"

        spans = [item.payload for item in items if item.type == "span"]
        span = None
        for sp in spans:
            if sp["attributes"]["sentry.op"].startswith("gen_ai"):
                assert span is None, "there is exactly one gen_ai span"
                span = sp
            else:
                # there should be no other spans, just the gen_ai span
                # and optionally some http.client spans from talking to the hf api
                assert sp["attributes"]["sentry.op"] == "http.client"

        assert span is not None
        assert span["status"] == "error"
    else:
        events = capture_events()

        with sentry_sdk.start_transaction(name="test"), pytest.raises(HfHubHTTPError):
            client.chat_completion(
                messages=[{"role": "user", "content": "Hello!"}],
            )

        (error, transaction) = events
        assert error["level"] == "error"

        span = None
        for sp in transaction["spans"]:
            if sp["op"].startswith("gen_ai"):
                assert span is None, "there is exactly one gen_ai span"
                span = sp
            else:
                # there should be no other spans, just the gen_ai span
                # and optionally some http.client spans from talking to the hf api
                assert sp["op"] == "http.client"

        assert span is not None
        assert span["status"] == "internal_error"
        assert span["tags"]["status"] == "internal_error"


@pytest.mark.parametrize("stream_gen_ai_spans", [True, False])
@pytest.mark.httpx_mock(assert_all_requests_were_expected=False)
@pytest.mark.parametrize("send_default_pii", [True, False])
@pytest.mark.parametrize("include_prompts", [True, False])
def test_chat_completion_with_tools(
    sentry_init: "Any",
    capture_events: "Any",
    capture_items: "Any",
    send_default_pii: "Any",
    include_prompts: "Any",
    mock_hf_chat_completion_api_tools: "Any",
    stream_gen_ai_spans: "Any",
):
    sentry_init(
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        integrations=[HuggingfaceHubIntegration(include_prompts=include_prompts)],
        stream_gen_ai_spans=stream_gen_ai_spans,
    )

    client = get_hf_provider_inference_client()

    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_weather",
                "description": "Get current weather",
                "parameters": {
                    "type": "object",
                    "properties": {"location": {"type": "string"}},
                    "required": ["location"],
                },
            },
        }
    ]

    if stream_gen_ai_spans:
        items = capture_items("transaction", "span")

        with sentry_sdk.start_transaction(name="test"):
            client.chat_completion(
                messages=[{"role": "user", "content": "What is the weather in Paris?"}],
                tools=tools,
                tool_choice="auto",
            )

        spans = [item.payload for item in items if item.type == "span"]
        span = None
        for sp in spans:
            if sp["attributes"]["sentry.op"].startswith("gen_ai"):
                assert span is None, "there is exactly one gen_ai span"
                span = sp
            else:
                # there should be no other spans, just the gen_ai span
                # and optionally some http.client spans from talking to the hf api
                assert sp["attributes"]["sentry.op"] == "http.client"

        assert span is not None

        assert span["attributes"]["sentry.op"] == "gen_ai.chat"
        assert span["name"] == "chat test-model"
        assert span["attributes"]["sentry.origin"] == "auto.ai.huggingface_hub"

        expected_data = {
            "gen_ai.operation.name": "chat",
            "gen_ai.request.available_tools": '[{"type": "function", "function": {"name": "get_weather", "description": "Get current weather", "parameters": {"type": "object", "properties": {"location": {"type": "string"}}, "required": ["location"]}}}]',
            "gen_ai.request.model": "test-model",
            "gen_ai.response.finish_reasons": "tool_calls",
            "gen_ai.response.model": "test-model-123",
            "gen_ai.usage.input_tokens": 10,
            "gen_ai.usage.output_tokens": 8,
            "gen_ai.usage.total_tokens": 18,
            "process.runtime.name": mock.ANY,
            "process.runtime.version": mock.ANY,
            "sentry.environment": "production",
            "sentry.op": "gen_ai.chat",
            "sentry.origin": "auto.ai.huggingface_hub",
            "sentry.release": mock.ANY,
            "sentry.sdk.name": "sentry.python",
            "sentry.sdk.version": mock.ANY,
            "sentry.segment.id": mock.ANY,
            "sentry.segment.name": "test",
            "server.address": mock.ANY,
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }

        if send_default_pii and include_prompts:
            expected_data["gen_ai.request.messages"] = (
                '[{"role": "user", "content": "What is the weather in Paris?"}]'
            )
            expected_data["gen_ai.response.tool_calls"] = (
                '[{"function": {"arguments": {"location": "Paris"}, "name": "get_weather", "description": "None"}, "id": "call_123", "type": "function"}]'
            )

        if not send_default_pii or not include_prompts:
            assert "gen_ai.request.messages" not in expected_data
            assert "gen_ai.response.text" not in expected_data
            assert "gen_ai.response.tool_calls" not in expected_data

        assert span["attributes"] == expected_data
    else:
        events = capture_events()

        with sentry_sdk.start_transaction(name="test"):
            client.chat_completion(
                messages=[{"role": "user", "content": "What is the weather in Paris?"}],
                tools=tools,
                tool_choice="auto",
            )

        (transaction,) = events

        span = None
        for sp in transaction["spans"]:
            if sp["op"].startswith("gen_ai"):
                assert span is None, "there is exactly one gen_ai span"
                span = sp
            else:
                # there should be no other spans, just the gen_ai span
                # and optionally some http.client spans from talking to the hf api
                assert sp["op"] == "http.client"

        assert span is not None

        assert span["op"] == "gen_ai.chat"
        assert span["description"] == "chat test-model"
        assert span["origin"] == "auto.ai.huggingface_hub"

        expected_data = {
            "gen_ai.operation.name": "chat",
            "gen_ai.request.available_tools": '[{"type": "function", "function": {"name": "get_weather", "description": "Get current weather", "parameters": {"type": "object", "properties": {"location": {"type": "string"}}, "required": ["location"]}}}]',
            "gen_ai.request.model": "test-model",
            "gen_ai.response.finish_reasons": "tool_calls",
            "gen_ai.response.model": "test-model-123",
            "gen_ai.usage.input_tokens": 10,
            "gen_ai.usage.output_tokens": 8,
            "gen_ai.usage.total_tokens": 18,
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }

        if send_default_pii and include_prompts:
            expected_data["gen_ai.request.messages"] = (
                '[{"role": "user", "content": "What is the weather in Paris?"}]'
            )
            expected_data["gen_ai.response.tool_calls"] = (
                '[{"function": {"arguments": {"location": "Paris"}, "name": "get_weather", "description": "None"}, "id": "call_123", "type": "function"}]'
            )

        if not send_default_pii or not include_prompts:
            assert "gen_ai.request.messages" not in expected_data
            assert "gen_ai.response.text" not in expected_data
            assert "gen_ai.response.tool_calls" not in expected_data

        assert span["data"] == expected_data


@pytest.mark.parametrize("stream_gen_ai_spans", [True, False])
@pytest.mark.httpx_mock(assert_all_requests_were_expected=False)
@pytest.mark.parametrize("send_default_pii", [True, False])
@pytest.mark.parametrize("include_prompts", [True, False])
def test_chat_completion_streaming_with_tools(
    sentry_init: "Any",
    capture_events: "Any",
    capture_items: "Any",
    send_default_pii: "Any",
    include_prompts: "Any",
    mock_hf_chat_completion_api_streaming_tools: "Any",
    stream_gen_ai_spans: "Any",
) -> None:
    sentry_init(
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        integrations=[HuggingfaceHubIntegration(include_prompts=include_prompts)],
        stream_gen_ai_spans=stream_gen_ai_spans,
    )

    client = get_hf_provider_inference_client()

    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_weather",
                "description": "Get current weather",
                "parameters": {
                    "type": "object",
                    "properties": {"location": {"type": "string"}},
                    "required": ["location"],
                },
            },
        }
    ]

    if stream_gen_ai_spans:
        items = capture_items("transaction", "span")

        with sentry_sdk.start_transaction(name="test"):
            _ = list(
                client.chat_completion(
                    messages=[
                        {"role": "user", "content": "What is the weather in Paris?"}
                    ],
                    stream=True,
                    tools=tools,
                    tool_choice="auto",
                )
            )

        spans = [item.payload for item in items if item.type == "span"]
        span = None
        for sp in spans:
            if sp["attributes"]["sentry.op"].startswith("gen_ai"):
                assert span is None, "there is exactly one gen_ai span"
                span = sp
            else:
                # there should be no other spans, just the gen_ai span
                # and optionally some http.client spans from talking to the hf api
                assert sp["attributes"]["sentry.op"] == "http.client"

        assert span is not None

        assert span["attributes"]["sentry.op"] == "gen_ai.chat"
        assert span["name"] == "chat test-model"
        assert span["attributes"]["sentry.origin"] == "auto.ai.huggingface_hub"

        expected_data = {
            "gen_ai.operation.name": "chat",
            "gen_ai.request.available_tools": '[{"type": "function", "function": {"name": "get_weather", "description": "Get current weather", "parameters": {"type": "object", "properties": {"location": {"type": "string"}}, "required": ["location"]}}}]',
            "gen_ai.request.model": "test-model",
            "gen_ai.response.finish_reasons": "tool_calls",
            "gen_ai.response.model": "test-model-123",
            "gen_ai.response.streaming": True,
            "process.runtime.name": mock.ANY,
            "process.runtime.version": mock.ANY,
            "sentry.environment": "production",
            "sentry.op": "gen_ai.chat",
            "sentry.origin": "auto.ai.huggingface_hub",
            "sentry.release": mock.ANY,
            "sentry.sdk.name": "sentry.python",
            "sentry.sdk.version": mock.ANY,
            "sentry.segment.id": mock.ANY,
            "sentry.segment.name": "test",
            "server.address": mock.ANY,
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }

        if HF_VERSION and HF_VERSION >= (0, 26, 0):
            expected_data["gen_ai.usage.input_tokens"] = 183
            expected_data["gen_ai.usage.output_tokens"] = 14
            expected_data["gen_ai.usage.total_tokens"] = 197

        if send_default_pii and include_prompts:
            expected_data["gen_ai.request.messages"] = (
                '[{"role": "user", "content": "What is the weather in Paris?"}]'
            )
            expected_data["gen_ai.response.text"] = "response with tool calls follows"
            expected_data["gen_ai.response.tool_calls"] = (
                '[{"function": {"arguments": {"location": "Paris"}, "name": "get_weather"}, "id": "call_123", "type": "function", "index": "None"}]'
            )

        if not send_default_pii or not include_prompts:
            assert "gen_ai.request.messages" not in expected_data
            assert "gen_ai.response.text" not in expected_data
            assert "gen_ai.response.tool_calls" not in expected_data

        assert span["attributes"] == expected_data
    else:
        events = capture_events()

        with sentry_sdk.start_transaction(name="test"):
            _ = list(
                client.chat_completion(
                    messages=[
                        {"role": "user", "content": "What is the weather in Paris?"}
                    ],
                    stream=True,
                    tools=tools,
                    tool_choice="auto",
                )
            )

        (transaction,) = events

        span = None
        for sp in transaction["spans"]:
            if sp["op"].startswith("gen_ai"):
                assert span is None, "there is exactly one gen_ai span"
                span = sp
            else:
                # there should be no other spans, just the gen_ai span
                # and optionally some http.client spans from talking to the hf api
                assert sp["op"] == "http.client"

        assert span is not None

        assert span["op"] == "gen_ai.chat"
        assert span["description"] == "chat test-model"
        assert span["origin"] == "auto.ai.huggingface_hub"

        expected_data = {
            "gen_ai.operation.name": "chat",
            "gen_ai.request.available_tools": '[{"type": "function", "function": {"name": "get_weather", "description": "Get current weather", "parameters": {"type": "object", "properties": {"location": {"type": "string"}}, "required": ["location"]}}}]',
            "gen_ai.request.model": "test-model",
            "gen_ai.response.finish_reasons": "tool_calls",
            "gen_ai.response.model": "test-model-123",
            "gen_ai.response.streaming": True,
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }

        if HF_VERSION and HF_VERSION >= (0, 26, 0):
            expected_data["gen_ai.usage.input_tokens"] = 183
            expected_data["gen_ai.usage.output_tokens"] = 14
            expected_data["gen_ai.usage.total_tokens"] = 197

        if send_default_pii and include_prompts:
            expected_data["gen_ai.request.messages"] = (
                '[{"role": "user", "content": "What is the weather in Paris?"}]'
            )
            expected_data["gen_ai.response.text"] = "response with tool calls follows"
            expected_data["gen_ai.response.tool_calls"] = (
                '[{"function": {"arguments": {"location": "Paris"}, "name": "get_weather"}, "id": "call_123", "type": "function", "index": "None"}]'
            )

        if not send_default_pii or not include_prompts:
            assert "gen_ai.request.messages" not in expected_data
            assert "gen_ai.response.text" not in expected_data
            assert "gen_ai.response.tool_calls" not in expected_data

        assert span["data"] == expected_data



import time

import pytest
from chalice import BadRequestError, Chalice
from chalice.local import LambdaContext, LocalGateway
from pytest_chalice.handlers import RequestHandler

import sentry_sdk
from sentry_sdk import capture_message
from sentry_sdk.integrations.chalice import CHALICE_VERSION, ChaliceIntegration
from sentry_sdk.utils import parse_version


def _populate_lambda_context(context):
    fn = context.function_name
    context.invoked_function_arn = (
        f"arn:aws:lambda:us-east-1:123456789012:function:{fn}"
    )
    context.log_group_name = f"/aws/lambda/{fn}"
    context.log_stream_name = "2024/01/01/[$LATEST]abcdef1234567890"
    context.aws_request_id = "test-request-id-1234"
    return context


def _generate_lambda_context(self):
    # Monkeypatch of the function _generate_lambda_context
    # from the class LocalGateway
    # for mock the timeout
    # type: () -> LambdaContext
    if self._config.lambda_timeout is None:
        timeout = 10 * 1000
    else:
        timeout = self._config.lambda_timeout * 1000
    context = LambdaContext(
        function_name=self._config.function_name,
        memory_size=self._config.lambda_memory_size,
        max_runtime_ms=timeout,
    )
    return _populate_lambda_context(context)


@pytest.fixture
def app(sentry_init):
    sentry_init(integrations=[ChaliceIntegration()])
    app = Chalice(app_name="sentry_chalice")

    @app.route("/boom")
    def boom():
        raise Exception("boom goes the dynamite!")

    @app.route("/context")
    def has_request():
        raise Exception("boom goes the dynamite!")

    @app.route("/badrequest")
    def badrequest():
        raise BadRequestError("bad-request")

    @app.route("/message")
    def hi():
        capture_message("hi")
        return {"status": "ok"}

    @app.route("/message/{message_id}")
    def hi_with_id(message_id):
        capture_message("hi again")
        return {"status": "ok"}

    LocalGateway._generate_lambda_context = _generate_lambda_context

    return app


@pytest.fixture
def lambda_context_args():
    return ["lambda_name", 256]


def test_exception_boom(app, client: RequestHandler) -> None:
    response = client.get("/boom")
    assert response.status_code == 500
    assert response.json == {
        "Code": "InternalServerError",
        "Message": "An internal server error occurred.",
    }


def test_has_request(app, capture_events, client: RequestHandler):
    events = capture_events()

    response = client.get("/context")
    assert response.status_code == 500

    (event,) = events
    assert event["level"] == "error"
    (exception,) = event["exception"]["values"]
    assert exception["type"] == "Exception"


def test_scheduled_event(app, lambda_context_args):
    @app.schedule("rate(1 minutes)")
    def every_hour(event):
        raise Exception("schedule event!")

    context = _populate_lambda_context(
        LambdaContext(*lambda_context_args, max_runtime_ms=10000, time_source=time)
    )

    lambda_event = {
        "version": "0",
        "account": "120987654312",
        "region": "us-west-1",
        "detail": {},
        "detail-type": "Scheduled Event",
        "source": "aws.events",
        "time": "1970-01-01T00:00:00Z",
        "id": "event-id",
        "resources": ["arn:aws:events:us-west-1:120987654312:rule/my-schedule"],
    }
    with pytest.raises(Exception) as exc_info:
        every_hour(lambda_event, context=context)
    assert str(exc_info.value) == "schedule event!"


@pytest.mark.skipif(
    parse_version(CHALICE_VERSION) >= (1, 26, 0),
    reason="different behavior based on chalice version",
)
def test_bad_request_old(client: RequestHandler) -> None:
    response = client.get("/badrequest")

    assert response.status_code == 400
    assert response.json == {
        "Code": "BadRequestError",
        "Message": "BadRequestError: bad-request",
    }


@pytest.mark.skipif(
    parse_version(CHALICE_VERSION) < (1, 26, 0),
    reason="different behavior based on chalice version",
)
def test_bad_request(client: RequestHandler) -> None:
    response = client.get("/badrequest")

    assert response.status_code == 400
    assert response.json == {
        "Code": "BadRequestError",
        "Message": "bad-request",
    }


@pytest.mark.parametrize(
    "url,expected_transaction,expected_source",
    [
        ("/message", "api_handler", "component"),
        ("/message/123456", "api_handler", "component"),
    ],
)
def test_transaction(
    app,
    client: RequestHandler,
    capture_events,
    url,
    expected_transaction,
    expected_source,
):
    events = capture_events()

    response = client.get(url)
    assert response.status_code == 200

    (event,) = events
    assert event["transaction"] == expected_transaction
    assert event["transaction_info"] == {"source": expected_source}


def _make_span_streaming_app(sentry_init):
    sentry_init(
        integrations=[ChaliceIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )
    app = Chalice(app_name="sentry_chalice")

    @app.route("/message")
    def hi():
        capture_message("hi")
        return {"status": "ok"}

    @app.route("/boom")
    def boom():
        raise Exception("boom goes the dynamite!")

    LocalGateway._generate_lambda_context = _generate_lambda_context

    return app


def test_span_streaming_existing_span(
    sentry_init,
    capture_items,
):
    """When a segment already exists (e.g. created by the AWS Lambda
    integration), Chalice decorates it instead of creating a duplicate."""
    app = _make_span_streaming_app(sentry_init)
    client = RequestHandler(app)
    items = capture_items("span")

    with sentry_sdk.traces.start_span(
        name="lambda_segment",
        parent_span=None,
        attributes={
            "sentry.origin": "auto.function.aws_lambda",
            "sentry.op": "function.aws",
            "faas.name": "api_handler",
        },
    ):
        response = client.get("/message")
        assert response.status_code == 200

    sentry_sdk.flush()

    segment_spans = [s.payload for s in items if s.payload.get("is_segment")]
    assert len(segment_spans) == 1
    span = segment_spans[0]

    attrs = span["attributes"]
    assert attrs["sentry.origin"] == "auto.function.chalice"
    assert attrs["sentry.op"] == "function.aws"
    assert attrs["faas.name"] == "api_handler"
    assert span["status"] == "ok"


def test_span_streaming_existing_span_error(
    sentry_init,
    capture_items,
):
    app = _make_span_streaming_app(sentry_init)
    client = RequestHandler(app)
    items = capture_items("event", "span")

    with sentry_sdk.traces.start_span(
        name="lambda_segment",
        parent_span=None,
        attributes={"sentry.origin": "auto.function.aws_lambda"},
    ):
        response = client.get("/boom")
        assert response.status_code == 500

    sentry_sdk.flush()

    error_items = [i for i in items if i.type == "event"]
    assert len(error_items) == 1

    segment_spans = [
        s.payload for s in items if s.type == "span" and s.payload.get("is_segment")
    ]
    assert len(segment_spans) == 1
    assert segment_spans[0]["attributes"]["sentry.origin"] == "auto.function.chalice"
    assert segment_spans[0]["status"] == "error"



from functools import wraps
from typing import TYPE_CHECKING

from django.dispatch import Signal

import sentry_sdk
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.integrations.django import DJANGO_VERSION
from sentry_sdk.tracing_utils import has_span_streaming_enabled

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any, Union


def _get_receiver_name(receiver: "Callable[..., Any]") -> str:
    name = ""

    if hasattr(receiver, "__qualname__"):
        name = receiver.__qualname__
    elif hasattr(receiver, "__name__"):  # Python 2.7 has no __qualname__
        name = receiver.__name__
    elif hasattr(
        receiver, "func"
    ):  # certain functions (like partials) dont have a name
        if hasattr(receiver, "func") and hasattr(receiver.func, "__name__"):
            name = "partial(<function " + receiver.func.__name__ + ">)"

    if (
        name == ""
    ):  # In case nothing was found, return the string representation (this is the slowest case)
        return str(receiver)

    if hasattr(receiver, "__module__"):  # prepend with module, if there is one
        name = receiver.__module__ + "." + name

    return name


def patch_signals() -> None:
    """
    Patch django signal receivers to create a span.

    This only wraps sync receivers. Django>=5.0 introduced async receivers, but
    since we don't create transactions for ASGI Django, we don't wrap them.
    """
    from sentry_sdk.integrations.django import DjangoIntegration

    old_live_receivers = Signal._live_receivers

    def _sentry_live_receivers(
        self: "Signal", sender: "Any"
    ) -> "Union[tuple[list[Callable[..., Any]], list[Callable[..., Any]]], list[Callable[..., Any]]]":
        if DJANGO_VERSION >= (5, 0):
            sync_receivers, async_receivers = old_live_receivers(self, sender)
        else:
            sync_receivers = old_live_receivers(self, sender)
            async_receivers = []

        def sentry_sync_receiver_wrapper(
            receiver: "Callable[..., Any]",
        ) -> "Callable[..., Any]":
            @wraps(receiver)
            def wrapper(*args: "Any", **kwargs: "Any") -> "Any":
                signal_name = _get_receiver_name(receiver)

                span_streaming = has_span_streaming_enabled(
                    sentry_sdk.get_client().options
                )
                if span_streaming:
                    if sentry_sdk.traces.get_current_span() is None:
                        return receiver(*args, **kwargs)
                    with sentry_sdk.traces.start_span(
                        name=signal_name,
                        attributes={
                            "sentry.op": OP.EVENT_DJANGO,
                            "sentry.origin": DjangoIntegration.origin,
                            SPANDATA.CODE_FUNCTION_NAME: signal_name,
                        },
                    ):
                        return receiver(*args, **kwargs)
                else:
                    with sentry_sdk.start_span(
                        op=OP.EVENT_DJANGO,
                        name=signal_name,
                        origin=DjangoIntegration.origin,
                    ) as span:
                        span.set_data("signal", signal_name)
                        return receiver(*args, **kwargs)

            return wrapper

        integration = sentry_sdk.get_client().get_integration(DjangoIntegration)
        if (
            integration
            and integration.signals_spans
            and self not in integration.signals_denylist
        ):
            for idx, receiver in enumerate(sync_receivers):
                sync_receivers[idx] = sentry_sync_receiver_wrapper(receiver)

        if DJANGO_VERSION >= (5, 0):
            return sync_receivers, async_receivers
        else:
            return sync_receivers

    Signal._live_receivers = _sentry_live_receivers



import json
import subprocess
import tempfile
import time
from unittest import mock

import boto3
import docker
import pytest
import yaml
from aws_cdk import App

from .utils import SAM_PORT, LocalLambdaStack, SentryServerForTesting

DOCKER_NETWORK_NAME = "lambda-test-network"
SAM_TEMPLATE_FILE = "sam.template.yaml"


@pytest.fixture(scope="session", autouse=True)
def test_environment():
    print("[test_environment fixture] Setting up AWS Lambda test infrastructure")

    # Create a Docker network
    docker_client = docker.from_env()
    docker_client.networks.prune()
    docker_client.networks.create(DOCKER_NETWORK_NAME, driver="bridge")

    # Start Sentry server
    server = SentryServerForTesting()
    server.start()
    time.sleep(1)  # Give it a moment to start up

    # Create local AWS SAM stack
    app = App()
    stack = LocalLambdaStack(app, "LocalLambdaStack")

    # Write SAM template to file
    template = app.synth().get_stack_by_name("LocalLambdaStack").template
    with open(SAM_TEMPLATE_FILE, "w") as f:
        yaml.dump(template, f)

    # Write SAM debug log to file
    debug_log_file = tempfile.gettempdir() + "/sentry_aws_lambda_tests_sam_debug.log"
    debug_log = open(debug_log_file, "w")
    print("[test_environment fixture] Writing SAM debug log to: %s" % debug_log_file)

    # Start SAM local
    process = subprocess.Popen(
        [
            "sam",
            "local",
            "start-lambda",
            "--debug",
            "--template",
            SAM_TEMPLATE_FILE,
            "--warm-containers",
            "EAGER",
            "--docker-network",
            DOCKER_NETWORK_NAME,
        ],
        stdout=debug_log,
        stderr=debug_log,
        text=True,  # This makes stdout/stderr return strings instead of bytes
    )

    try:
        # Wait for SAM to be ready
        LocalLambdaStack.wait_for_stack()

        def before_test():
            server.clear_envelopes()

        yield {
            "stack": stack,
            "server": server,
            "before_test": before_test,
        }

    finally:
        print("[test_environment fixture] Tearing down AWS Lambda test infrastructure")

        process.terminate()
        process.wait(timeout=10)  # Give it time to shut down gracefully

        # Force kill if still running
        if process.poll() is None:
            process.kill()


@pytest.fixture(autouse=True)
def clear_before_test(test_environment):
    test_environment["before_test"]()


@pytest.fixture
def lambda_client():
    """
    Create a boto3 client configured to use the local AWS SAM instance.
    """
    return boto3.client(
        "lambda",
        endpoint_url=f"http://127.0.0.1:{SAM_PORT}",  # noqa: E231
        aws_access_key_id="dummy",
        aws_secret_access_key="dummy",
        region_name="us-east-1",
    )


def test_basic_no_exception(lambda_client, test_environment):
    lambda_client.invoke(
        FunctionName="BasicOk",
        Payload=json.dumps({}),
    )
    envelopes = test_environment["server"].envelopes

    (transaction_event,) = envelopes

    assert transaction_event["type"] == "transaction"
    assert transaction_event["transaction"] == "BasicOk"
    assert transaction_event["sdk"]["name"] == "sentry.python.aws_lambda"
    assert transaction_event["tags"] == {"aws_region": "us-east-1"}

    assert transaction_event["extra"]["cloudwatch logs"] == {
        "log_group": mock.ANY,
        "log_stream": mock.ANY,
        "url": mock.ANY,
    }
    assert transaction_event["extra"]["lambda"] == {
        "aws_request_id": mock.ANY,
        "execution_duration_in_millis": mock.ANY,
        "function_name": "BasicOk",
        "function_version": "$LATEST",
        "invoked_function_arn": "arn:aws:lambda:us-east-1:012345678912:function:BasicOk",
        "remaining_time_in_millis": mock.ANY,
    }
    assert transaction_event["contexts"]["trace"] == {
        "op": "function.aws",
        "description": mock.ANY,
        "span_id": mock.ANY,
        "parent_span_id": mock.ANY,
        "trace_id": mock.ANY,
        "origin": "auto.function.aws_lambda",
        "data": mock.ANY,
    }


def test_basic_exception(lambda_client, test_environment):
    lambda_client.invoke(
        FunctionName="BasicException",
        Payload=json.dumps({}),
    )
    envelopes = test_environment["server"].envelopes

    # The second envelope we ignore.
    # It is the transaction that we test in test_basic_no_exception.
    (error_event, _) = envelopes

    assert error_event["level"] == "error"
    assert error_event["exception"]["values"][0]["type"] == "RuntimeError"
    assert error_event["exception"]["values"][0]["value"] == "Oh!"
    assert error_event["sdk"]["name"] == "sentry.python.aws_lambda"

    assert error_event["tags"] == {"aws_region": "us-east-1"}
    assert error_event["extra"]["cloudwatch logs"] == {
        "log_group": mock.ANY,
        "log_stream": mock.ANY,
        "url": mock.ANY,
    }
    assert error_event["extra"]["lambda"] == {
        "aws_request_id": mock.ANY,
        "execution_duration_in_millis": mock.ANY,
        "function_name": "BasicException",
        "function_version": "$LATEST",
        "invoked_function_arn": "arn:aws:lambda:us-east-1:012345678912:function:BasicException",
        "remaining_time_in_millis": mock.ANY,
    }
    assert error_event["contexts"]["trace"] == {
        "op": "function.aws",
        "description": mock.ANY,
        "span_id": mock.ANY,
        "parent_span_id": mock.ANY,
        "trace_id": mock.ANY,
        "origin": "auto.function.aws_lambda",
        "data": mock.ANY,
    }


def test_init_error(lambda_client, test_environment):
    lambda_client.invoke(
        FunctionName="InitError",
        Payload=json.dumps({}),
    )
    envelopes = test_environment["server"].envelopes

    (error_event, transaction_event) = envelopes

    assert (
        error_event["exception"]["values"][0]["value"] == "name 'func' is not defined"
    )
    assert transaction_event["transaction"] == "InitError"


def test_timeout_error(lambda_client, test_environment):
    lambda_client.invoke(
        FunctionName="TimeoutError",
        Payload=json.dumps({}),
    )
    envelopes = test_environment["server"].envelopes

    (error_event,) = envelopes

    assert error_event["level"] == "error"
    assert error_event["extra"]["lambda"]["function_name"] == "TimeoutError"

    (exception,) = error_event["exception"]["values"]
    assert not exception["mechanism"]["handled"]
    assert exception["type"] == "ServerlessTimeoutWarning"
    assert exception["value"].startswith(
        "WARNING : Function is expected to get timed out. Configured timeout duration ="
    )
    assert exception["mechanism"]["type"] == "threading"


def test_timeout_error_scope_modified(lambda_client, test_environment):
    lambda_client.invoke(
        FunctionName="TimeoutErrorScopeModified",
        Payload=json.dumps({}),
    )
    envelopes = test_environment["server"].envelopes

    (error_event,) = envelopes

    assert error_event["level"] == "error"
    assert (
        error_event["extra"]["lambda"]["function_name"] == "TimeoutErrorScopeModified"
    )

    (exception,) = error_event["exception"]["values"]
    assert not exception["mechanism"]["handled"]
    assert exception["type"] == "ServerlessTimeoutWarning"
    assert exception["value"].startswith(
        "WARNING : Function is expected to get timed out. Configured timeout duration ="
    )
    assert exception["mechanism"]["type"] == "threading"

    assert error_event["tags"]["custom_tag"] == "custom_value"


@pytest.mark.parametrize(
    "aws_event, has_request_data, batch_size",
    [
        (b"1231", False, 1),
        (b"11.21", False, 1),
        (b'"Good dog!"', False, 1),
        (b"true", False, 1),
        (
            b"""
            [
                {"good dog": "Maisey"},
                {"good dog": "Charlie"},
                {"good dog": "Cory"},
                {"good dog": "Bodhi"}
            ]
            """,
            False,
            4,
        ),
        (
            b"""
            [
                {
                    "headers": {
                        "Host": "x1.io",
                        "X-Forwarded-Proto": "https"
                    },
                    "httpMethod": "GET",
                    "path": "/1",
                    "queryStringParameters": {
                        "done": "f"
                    },
                    "d": "D1"
                },
                {
                    "headers": {
                        "Host": "x2.io",
                        "X-Forwarded-Proto": "http"
                    },
                    "httpMethod": "POST",
                    "path": "/2",
                    "queryStringParameters": {
                        "done": "t"
                    },
                    "d": "D2"
                }
            ]
            """,
            True,
            2,
        ),
        (b"[]", False, 1),
    ],
    ids=[
        "event as integer",
        "event as float",
        "event as string",
        "event as bool",
        "event as list of dicts",
        "event as dict",
        "event as empty list",
    ],
)
def test_non_dict_event(
    lambda_client, test_environment, aws_event, has_request_data, batch_size
):
    lambda_client.invoke(
        FunctionName="BasicException",
        Payload=aws_event,
    )
    envelopes = test_environment["server"].envelopes

    (error_event, transaction_event) = envelopes

    assert transaction_event["type"] == "transaction"
    assert transaction_event["transaction"] == "BasicException"
    assert transaction_event["sdk"]["name"] == "sentry.python.aws_lambda"
    assert transaction_event["contexts"]["trace"]["status"] == "internal_error"

    assert error_event["level"] == "error"
    assert error_event["transaction"] == "BasicException"
    assert error_event["sdk"]["name"] == "sentry.python.aws_lambda"
    assert error_event["exception"]["values"][0]["type"] == "RuntimeError"
    assert error_event["exception"]["values"][0]["value"] == "Oh!"
    assert error_event["exception"]["values"][0]["mechanism"]["type"] == "aws_lambda"

    if has_request_data:
        request_data = {
            "headers": {"Host": "x1.io", "X-Forwarded-Proto": "https"},
            "method": "GET",
            "url": "https://x1.io/1",
            "query_string": {
                "done": "f",
            },
        }
    else:
        request_data = {"url": "awslambda:///BasicException"}

    assert error_event["request"] == request_data
    assert transaction_event["request"] == request_data

    if batch_size > 1:
        assert error_event["tags"]["batch_size"] == batch_size
        assert error_event["tags"]["batch_request"] is True
        assert transaction_event["tags"]["batch_size"] == batch_size
        assert transaction_event["tags"]["batch_request"] is True


def test_request_data(lambda_client, test_environment):
    payload = b"""
        {
          "resource": "/asd",
          "path": "/asd",
          "httpMethod": "GET",
          "headers": {
            "Host": "iwsz2c7uwi.execute-api.us-east-1.amazonaws.com",
            "User-Agent": "custom",
            "X-Forwarded-Proto": "https"
          },
          "queryStringParameters": {
            "bonkers": "true"
          },
          "pathParameters": null,
          "stageVariables": null,
          "requestContext": {
            "identity": {
              "sourceIp": "213.47.147.207",
              "userArn": "42"
            }
          },
          "body": null,
          "isBase64Encoded": false
        }
    """

    lambda_client.invoke(
        FunctionName="BasicOk",
        Payload=payload,
    )
    envelopes = test_environment["server"].envelopes

    (transaction_event,) = envelopes

    assert transaction_event["request"] == {
        "headers": {
            "Host": "iwsz2c7uwi.execute-api.us-east-1.amazonaws.com",
            "User-Agent": "custom",
            "X-Forwarded-Proto": "https",
        },
        "method": "GET",
        "query_string": {"bonkers": "true"},
        "url": "https://iwsz2c7uwi.execute-api.us-east-1.amazonaws.com/asd",
    }


def test_trace_continuation(lambda_client, test_environment):
    trace_id = "471a43a4192642f0b136d5159a501701"
    parent_span_id = "6e8f22c393e68f19"
    parent_sampled = 1
    sentry_trace_header = "{}-{}-{}".format(trace_id, parent_span_id, parent_sampled)

    # We simulate here AWS Api Gateway's behavior of passing HTTP headers
    # as the `headers` dict in the event passed to the Lambda function.
    payload = {
        "headers": {
            "sentry-trace": sentry_trace_header,
        }
    }

    lambda_client.invoke(
        FunctionName="BasicException",
        Payload=json.dumps(payload),
    )
    envelopes = test_environment["server"].envelopes

    (error_event, transaction_event) = envelopes

    assert (
        error_event["contexts"]["trace"]["trace_id"]
        == transaction_event["contexts"]["trace"]["trace_id"]
        == "471a43a4192642f0b136d5159a501701"
    )


@pytest.mark.parametrize(
    "payload",
    [
        {},
        {"headers": None},
        {"headers": ""},
        {"headers": {}},
        {"headers": []},  # EventBridge sends an empty list
    ],
    ids=[
        "no headers",
        "none headers",
        "empty string headers",
        "empty dict headers",
        "empty list headers",
    ],
)
def test_headers(lambda_client, test_environment, payload):
    lambda_client.invoke(
        FunctionName="BasicException",
        Payload=json.dumps(payload),
    )
    envelopes = test_environment["server"].envelopes

    (error_event, _) = envelopes

    assert error_event["level"] == "error"
    assert error_event["exception"]["values"][0]["type"] == "RuntimeError"
    assert error_event["exception"]["values"][0]["value"] == "Oh!"


def test_span_origin(lambda_client, test_environment):
    lambda_client.invoke(
        FunctionName="BasicOk",
        Payload=json.dumps({}),
    )
    envelopes = test_environment["server"].envelopes

    (transaction_event,) = envelopes

    assert (
        transaction_event["contexts"]["trace"]["origin"] == "auto.function.aws_lambda"
    )


def test_traces_sampler_has_correct_sampling_context(lambda_client, test_environment):
    """
    Test that aws_event and aws_context are passed in the custom_sampling_context
    when using the AWS Lambda integration.
    """
    test_payload = {"test_key": "test_value"}
    response = lambda_client.invoke(
        FunctionName="TracesSampler",
        Payload=json.dumps(test_payload),
    )
    response_payload = json.loads(response["Payload"].read().decode())
    sampling_context_data = json.loads(response_payload["body"])[
        "sampling_context_data"
    ]
    assert sampling_context_data.get("aws_event_present") is True
    assert sampling_context_data.get("aws_context_present") is True
    assert sampling_context_data.get("event_data", {}).get("test_key") == "test_value"


@pytest.mark.parametrize(
    "lambda_function_name",
    ["RaiseErrorPerformanceEnabled", "RaiseErrorPerformanceDisabled"],
)
def test_error_has_new_trace_context(
    lambda_client, test_environment, lambda_function_name
):
    lambda_client.invoke(
        FunctionName=lambda_function_name,
        Payload=json.dumps({}),
    )
    envelopes = test_environment["server"].envelopes

    if lambda_function_name == "RaiseErrorPerformanceEnabled":
        (error_event, transaction_event) = envelopes
    else:
        (error_event,) = envelopes
        transaction_event = None

    assert "trace" in error_event["contexts"]
    assert "trace_id" in error_event["contexts"]["trace"]

    if transaction_event:
        assert "trace" in transaction_event["contexts"]
        assert "trace_id" in transaction_event["contexts"]["trace"]
        assert (
            error_event["contexts"]["trace"]["trace_id"]
            == transaction_event["contexts"]["trace"]["trace_id"]
        )


def _get_span_attr(attrs, key):
    """Extract the value from a span attribute, handling both flat and typed formats."""
    val = attrs[key]
    if isinstance(val, dict) and "value" in val:
        return val["value"]
    return val


def test_span_streaming_no_error(lambda_client, test_environment):
    lambda_client.invoke(
        FunctionName="BasicOkSpanStreaming",
        Payload=json.dumps({}),
    )
    envelopes = test_environment["server"].envelopes
    span_items = test_environment["server"].span_items

    assert len(envelopes) == 0

    segment_spans = [s for s in span_items if s["is_segment"]]
    assert len(segment_spans) == 1
    segment_span = segment_spans[0]

    assert segment_span["name"] == "BasicOkSpanStreaming"

    attrs = segment_span["attributes"]

    assert _get_span_attr(attrs, "sentry.op") == "function.aws"
    assert _get_span_attr(attrs, "sentry.origin") == "auto.function.aws_lambda"
    assert _get_span_attr(attrs, "sentry.segment.name.source") == "component"
    assert _get_span_attr(attrs, "cloud.provider") == "aws"
    assert _get_span_attr(attrs, "cloud.platform") == "aws_lambda"
    assert (
        _get_span_attr(attrs, "cloud.resource_id")
        == "arn:aws:lambda:us-east-1:012345678912:function:BasicOkSpanStreaming"
    )
    assert _get_span_attr(attrs, "cloud.region") == "us-east-1"
    assert _get_span_attr(attrs, "faas.name") == "BasicOkSpanStreaming"
    assert _get_span_attr(attrs, "faas.version") == "$LATEST"
    assert "faas.invocation_id" in attrs
    assert (
        _get_span_attr(attrs, "aws.lambda.invoked_arn")
        == "arn:aws:lambda:us-east-1:012345678912:function:BasicOkSpanStreaming"
    )
    assert _get_span_attr(attrs, "aws.log.group.names") == [
        "aws/lambda/BasicOkSpanStreaming"
    ]
    assert _get_span_attr(attrs, "aws.log.stream.names") == ["$LATEST"]
    assert _get_span_attr(attrs, "messaging.batch.message_count") == 1


def test_span_streaming_error(lambda_client, test_environment):
    lambda_client.invoke(
        FunctionName="RaiseErrorSpanStreaming",
        Payload=json.dumps({}),
    )
    envelopes = test_environment["server"].envelopes
    span_items = test_environment["server"].span_items

    assert len(envelopes) == 1
    error_event = envelopes[0]
    assert error_event["level"] == "error"
    (exception,) = error_event["exception"]["values"]
    assert exception["type"] == "Exception"
    assert exception["value"] == "Oh!"
    assert exception["mechanism"]["type"] == "aws_lambda"
    assert not exception["mechanism"]["handled"]

    segment_spans = [s for s in span_items if s["is_segment"]]
    assert len(segment_spans) == 1
    segment_span = segment_spans[0]

    assert segment_span["name"] == "RaiseErrorSpanStreaming"
    assert segment_span["status"] == "error"

    attrs = segment_span["attributes"]

    assert _get_span_attr(attrs, "sentry.op") == "function.aws"
    assert _get_span_attr(attrs, "sentry.origin") == "auto.function.aws_lambda"
    assert _get_span_attr(attrs, "sentry.segment.name.source") == "component"
    assert _get_span_attr(attrs, "cloud.provider") == "aws"
    assert _get_span_attr(attrs, "cloud.platform") == "aws_lambda"
    assert (
        _get_span_attr(attrs, "cloud.resource_id")
        == "arn:aws:lambda:us-east-1:012345678912:function:RaiseErrorSpanStreaming"
    )
    assert _get_span_attr(attrs, "cloud.region") == "us-east-1"
    assert _get_span_attr(attrs, "faas.name") == "RaiseErrorSpanStreaming"
    assert _get_span_attr(attrs, "faas.version") == "$LATEST"
    assert "faas.invocation_id" in attrs
    assert (
        _get_span_attr(attrs, "aws.lambda.invoked_arn")
        == "arn:aws:lambda:us-east-1:012345678912:function:RaiseErrorSpanStreaming"
    )
    assert _get_span_attr(attrs, "aws.log.group.names") == [
        "aws/lambda/RaiseErrorSpanStreaming"
    ]
    assert _get_span_attr(attrs, "aws.log.stream.names") == ["$LATEST"]
    assert _get_span_attr(attrs, "messaging.batch.message_count") == 1


def test_span_streaming_trace_continuation(lambda_client, test_environment):
    trace_id = "471a43a4192642f0b136d5159a501701"
    parent_span_id = "6e8f22c393e68f19"
    parent_sampled = 1
    sentry_trace_header = "{}-{}-{}".format(trace_id, parent_span_id, parent_sampled)

    payload = {
        "headers": {
            "sentry-trace": sentry_trace_header,
        }
    }

    lambda_client.invoke(
        FunctionName="RaiseErrorSpanStreaming",
        Payload=json.dumps(payload),
    )
    envelopes = test_environment["server"].envelopes
    span_items = test_environment["server"].span_items

    assert len(envelopes) == 1
    error_event = envelopes[0]
    assert error_event["contexts"]["trace"]["trace_id"] == trace_id

    segment_spans = [s for s in span_items if s["is_segment"]]
    assert len(segment_spans) == 1
    segment_span = segment_spans[0]
    assert segment_span["trace_id"] == trace_id
    assert segment_span["name"] == "RaiseErrorSpanStreaming"
    attrs = segment_span["attributes"]
    assert _get_span_attr(attrs, "sentry.op") == "function.aws"
    assert _get_span_attr(attrs, "sentry.origin") == "auto.function.aws_lambda"
    assert _get_span_attr(attrs, "sentry.segment.name.source") == "component"
    assert _get_span_attr(attrs, "cloud.provider") == "aws"
    assert _get_span_attr(attrs, "cloud.platform") == "aws_lambda"
    assert _get_span_attr(attrs, "cloud.region") == "us-east-1"
    assert _get_span_attr(attrs, "faas.name") == "RaiseErrorSpanStreaming"
    assert _get_span_attr(attrs, "faas.version") == "$LATEST"
    assert "faas.invocation_id" in attrs


def test_span_streaming_request_attributes(lambda_client, test_environment):
    payload = {
        "headers": {
            "Content-Type": "application/json",
            "Accept": "text/html",
        },
        "httpMethod": "POST",
        "queryStringParameters": {"foo": "bar", "a-complicated-value": "a=b&c=d"},
        "path": "/test",
    }

    lambda_client.invoke(
        FunctionName="BasicOkSpanStreamingPii",
        Payload=json.dumps(payload),
    )
    span_items = test_environment["server"].span_items

    segment_spans = [s for s in span_items if s["is_segment"]]
    assert len(segment_spans) == 1
    segment_span = segment_spans[0]
    attrs = segment_span["attributes"]

    assert _get_span_attr(attrs, "http.request.method") == "POST"
    assert (
        _get_span_attr(attrs, "url.query")
        == "foo=bar&a-complicated-value=a%3Db%26c%3Dd"
    )
    assert (
        _get_span_attr(attrs, "http.request.header.content-type") == "application/json"
    )
    assert _get_span_attr(attrs, "http.request.header.accept") == "text/html"
    assert _get_span_attr(attrs, "faas.name") == "BasicOkSpanStreamingPii"
    assert _get_span_attr(attrs, "cloud.provider") == "aws"
    assert _get_span_attr(attrs, "cloud.platform") == "aws_lambda"
    assert _get_span_attr(attrs, "cloud.region") == "us-east-1"
    assert _get_span_attr(attrs, "faas.version") == "$LATEST"
    assert "faas.invocation_id" in attrs
    assert _get_span_attr(attrs, "aws.log.group.names") == [
        "aws/lambda/BasicOkSpanStreamingPii"
    ]
    assert _get_span_attr(attrs, "aws.log.stream.names") == ["$LATEST"]


@pytest.mark.parametrize(
    "lambda_function_name",
    ["RaiseErrorPerformanceEnabled", "RaiseErrorPerformanceDisabled"],
)
def test_error_has_existing_trace_context(
    lambda_client, test_environment, lambda_function_name
):
    trace_id = "471a43a4192642f0b136d5159a501701"
    parent_span_id = "6e8f22c393e68f19"
    parent_sampled = 1
    sentry_trace_header = "{}-{}-{}".format(trace_id, parent_span_id, parent_sampled)

    # We simulate here AWS Api Gateway's behavior of passing HTTP headers
    # as the `headers` dict in the event passed to the Lambda function.
    payload = {
        "headers": {
            "sentry-trace": sentry_trace_header,
        }
    }

    lambda_client.invoke(
        FunctionName=lambda_function_name,
        Payload=json.dumps(payload),
    )
    envelopes = test_environment["server"].envelopes

    if lambda_function_name == "RaiseErrorPerformanceEnabled":
        (error_event, transaction_event) = envelopes
    else:
        (error_event,) = envelopes
        transaction_event = None

    assert "trace" in error_event["contexts"]
    assert "trace_id" in error_event["contexts"]["trace"]
    assert (
        error_event["contexts"]["trace"]["trace_id"]
        == "471a43a4192642f0b136d5159a501701"
    )

    if transaction_event:
        assert "trace" in transaction_event["contexts"]
        assert "trace_id" in transaction_event["contexts"]["trace"]
        assert (
            transaction_event["contexts"]["trace"]["trace_id"]
            == "471a43a4192642f0b136d5159a501701"
        )



import logging

import falcon
import falcon.testing
import pytest

import sentry_sdk
from sentry_sdk.integrations.falcon import FalconIntegration
from sentry_sdk.integrations.logging import LoggingIntegration
from sentry_sdk.utils import parse_version

try:
    import falcon.asgi
except ImportError:
    pass
else:
    import falcon.inspect  # We only need this module for the ASGI test


FALCON_VERSION = parse_version(falcon.__version__)


@pytest.fixture
def make_app(sentry_init):
    def inner():
        class MessageResource:
            def on_get(self, req, resp):
                sentry_sdk.capture_message("hi")
                resp.media = "hi"

        class MessageByIdResource:
            def on_get(self, req, resp, message_id):
                sentry_sdk.capture_message("hi")
                resp.media = "hi"

        class CustomError(Exception):
            pass

        class CustomErrorResource:
            def on_get(self, req, resp):
                raise CustomError()

        def custom_error_handler(*args, **kwargs):
            raise falcon.HTTPError(status=falcon.HTTP_400)

        app = falcon.API()
        app.add_route("/message", MessageResource())
        app.add_route("/message/{message_id:int}", MessageByIdResource())
        app.add_route("/custom-error", CustomErrorResource())

        app.add_error_handler(CustomError, custom_error_handler)

        return app

    return inner


@pytest.fixture
def make_client(make_app):
    def inner():
        app = make_app()
        return falcon.testing.TestClient(app)

    return inner


@pytest.mark.parametrize("span_streaming", [True, False])
def test_has_context(
    sentry_init,
    capture_events,
    capture_items,
    make_client,
    span_streaming,
):
    sentry_init(
        integrations=[FalconIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    client = make_client()
    if span_streaming:
        items = capture_items("event")

        response = client.simulate_get("/message")
        assert response.status == falcon.HTTP_200

        (event,) = (item.payload for item in items)
    else:
        events = capture_events()

        response = client.simulate_get("/message")
        assert response.status == falcon.HTTP_200

        (event,) = events
    assert event["transaction"] == "/message"  # Falcon URI template
    assert "data" not in event["request"]
    assert event["request"]["url"] == "http://falconframework.org/message"


@pytest.mark.parametrize(
    "url,transaction_style,expected_transaction,expected_source",
    [
        ("/message", "uri_template", "/message", "route"),
        ("/message", "path", "/message", "url"),
        ("/message/123456", "uri_template", "/message/{message_id:int}", "route"),
        ("/message/123456", "path", "/message/123456", "url"),
    ],
)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_transaction_style(
    sentry_init,
    make_client,
    capture_events,
    capture_items,
    url,
    transaction_style,
    expected_transaction,
    expected_source,
    span_streaming,
):
    integration = FalconIntegration(transaction_style=transaction_style)
    sentry_init(
        integrations=[integration],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    client = make_client()
    if span_streaming:
        items = capture_items("event")
        items = capture_items("event", "span")

        response = client.simulate_get(url)
        assert response.status == falcon.HTTP_200

        (event,) = (item.payload for item in items if item.type == "event")

        sentry_sdk.flush()
        spans = [item.payload for item in items if item.type == "span"]
        spans = [span for span in spans if span["name"] == expected_transaction]
        assert len(spans) == 1
        assert spans[0]["attributes"]["sentry.segment.name.source"] == expected_source
    else:
        events = capture_events()

        response = client.simulate_get(url)
        assert response.status == falcon.HTTP_200

        (event, transaction) = events

        assert transaction["transaction"] == expected_transaction
        assert transaction["transaction_info"] == {"source": expected_source}

    assert event["transaction"] == expected_transaction
    assert event["transaction_info"] == {"source": expected_source}


@pytest.mark.parametrize("span_streaming", [True, False])
def test_unhandled_errors(
    sentry_init,
    capture_exceptions,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[FalconIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    class Resource:
        def on_get(self, req, resp):
            1 / 0

    app = falcon.API()
    app.add_route("/", Resource())

    client = falcon.testing.TestClient(app)

    exceptions = capture_exceptions()
    if span_streaming:
        items = capture_items("event")

        try:
            client.simulate_get("/")
        except ZeroDivisionError:
            pass

        (exc,) = exceptions
        assert isinstance(exc, ZeroDivisionError)

        (event,) = (item.payload for item in items)
    else:
        events = capture_events()

        try:
            client.simulate_get("/")
        except ZeroDivisionError:
            pass

        (exc,) = exceptions
        assert isinstance(exc, ZeroDivisionError)

        (event,) = events
    assert event["exception"]["values"][0]["mechanism"]["type"] == "falcon"
    assert " by zero" in event["exception"]["values"][0]["value"]


@pytest.mark.parametrize("span_streaming", [True, False])
def test_raised_5xx_errors(
    sentry_init,
    capture_exceptions,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[FalconIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    class Resource:
        def on_get(self, req, resp):
            raise falcon.HTTPError(falcon.HTTP_502)

    app = falcon.API()
    app.add_route("/", Resource())

    client = falcon.testing.TestClient(app)

    exceptions = capture_exceptions()
    if span_streaming:
        items = capture_items("event")

        client.simulate_get("/")

        (exc,) = exceptions
        assert isinstance(exc, falcon.HTTPError)

        (event,) = (item.payload for item in items)
    else:
        events = capture_events()

        client.simulate_get("/")

        (exc,) = exceptions
        assert isinstance(exc, falcon.HTTPError)

        (event,) = events
    assert event["exception"]["values"][0]["mechanism"]["type"] == "falcon"
    assert event["exception"]["values"][0]["type"] == "HTTPError"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_raised_4xx_errors(
    sentry_init,
    capture_exceptions,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[FalconIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    class Resource:
        def on_get(self, req, resp):
            raise falcon.HTTPError(falcon.HTTP_400)

    app = falcon.API()
    app.add_route("/", Resource())

    exceptions = capture_exceptions()
    if span_streaming:
        events = capture_items("event")
    else:
        events = capture_events()

    client = falcon.testing.TestClient(app)
    client.simulate_get("/")

    assert len(exceptions) == 0
    assert len(events) == 0


@pytest.mark.parametrize("span_streaming", [True, False])
def test_http_status(
    sentry_init,
    capture_exceptions,
    capture_events,
    capture_items,
    span_streaming,
):
    """
    This just demonstrates, that if Falcon raises a HTTPStatus with code 500
    (instead of a HTTPError with code 500) Sentry will not capture it.
    """
    sentry_init(
        integrations=[FalconIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    class Resource:
        def on_get(self, req, resp):
            raise falcon.http_status.HTTPStatus(falcon.HTTP_508)

    app = falcon.API()
    app.add_route("/", Resource())

    client = falcon.testing.TestClient(app)

    exceptions = capture_exceptions()
    if span_streaming:
        events = capture_items("event")

        client.simulate_get("/")
    else:
        events = capture_events()

        client.simulate_get("/")

    assert len(exceptions) == 0
    assert len(events) == 0


@pytest.mark.parametrize("max_value_length", [1024, None])
@pytest.mark.parametrize("span_streaming", [True, False])
def test_falcon_large_json_request(
    sentry_init,
    capture_events,
    capture_items,
    max_value_length,
    span_streaming,
):
    sentry_init(
        integrations=[FalconIntegration()],
        max_request_body_size="always",
        max_value_length=max_value_length,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    data = {"foo": {"bar": "a" * (1034)}}

    class Resource:
        def on_post(self, req, resp):
            assert req.media == data
            sentry_sdk.capture_message("hi")
            resp.media = "ok"

    app = falcon.API()
    app.add_route("/", Resource())

    client = falcon.testing.TestClient(app)

    if span_streaming:
        items = capture_items("event")

        response = client.simulate_post("/", json=data)
        assert response.status == falcon.HTTP_200

        (event,) = (item.payload for item in items)
    else:
        events = capture_events()

        response = client.simulate_post("/", json=data)
        assert response.status == falcon.HTTP_200

        (event,) = events
    if max_value_length:
        assert event["_meta"]["request"]["data"]["foo"]["bar"] == {
            "": {
                "len": 1034,
                "rem": [["!limit", "x", 1021, 1024]],
            }
        }
        assert len(event["request"]["data"]["foo"]["bar"]) == 1024
    else:
        assert len(event["request"]["data"]["foo"]["bar"]) == 1034


@pytest.mark.parametrize("data", [{}, []], ids=["empty-dict", "empty-list"])
@pytest.mark.parametrize("span_streaming", [True, False])
def test_falcon_empty_json_request(
    sentry_init,
    capture_events,
    capture_items,
    data,
    span_streaming,
):
    sentry_init(
        integrations=[FalconIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    class Resource:
        def on_post(self, req, resp):
            assert req.media == data
            sentry_sdk.capture_message("hi")
            resp.media = "ok"

    app = falcon.API()
    app.add_route("/", Resource())

    client = falcon.testing.TestClient(app)

    if span_streaming:
        items = capture_items("event")

        response = client.simulate_post("/", json=data)
        assert response.status == falcon.HTTP_200

        (event,) = (item.payload for item in items)
    else:
        events = capture_events()

        response = client.simulate_post("/", json=data)
        assert response.status == falcon.HTTP_200

        (event,) = events
    assert event["request"]["data"] == data


@pytest.mark.parametrize("span_streaming", [True, False])
def test_falcon_raw_data_request(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[FalconIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    class Resource:
        def on_post(self, req, resp):
            sentry_sdk.capture_message("hi")
            resp.media = "ok"

    app = falcon.API()
    app.add_route("/", Resource())

    client = falcon.testing.TestClient(app)

    if span_streaming:
        items = capture_items("event")

        response = client.simulate_post("/", body="hi")
        assert response.status == falcon.HTTP_200

        (event,) = (item.payload for item in items)
    else:
        events = capture_events()

        response = client.simulate_post("/", body="hi")
        assert response.status == falcon.HTTP_200

        (event,) = events
    assert event["request"]["headers"]["Content-Length"] == "2"
    assert event["request"]["data"] == ""


@pytest.mark.parametrize("span_streaming", [True, False])
def test_logging(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[FalconIntegration(), LoggingIntegration(event_level="ERROR")],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    logger = logging.getLogger()

    app = falcon.API()

    class Resource:
        def on_get(self, req, resp):
            logger.error("hi")
            resp.media = "ok"

    app.add_route("/", Resource())

    client = falcon.testing.TestClient(app)

    if span_streaming:
        items = capture_items("event")

        client.simulate_get("/")

        (event,) = (item.payload for item in items)
    else:
        events = capture_events()

        client.simulate_get("/")

        (event,) = events
    assert event["level"] == "error"


def test_500(sentry_init):
    sentry_init(integrations=[FalconIntegration()])

    app = falcon.API()

    class Resource:
        def on_get(self, req, resp):
            1 / 0

    app.add_route("/", Resource())

    def http500_handler(ex, req, resp, params):
        sentry_sdk.capture_exception(ex)
        resp.media = {"message": "Sentry error."}

    app.add_error_handler(Exception, http500_handler)

    client = falcon.testing.TestClient(app)
    response = client.simulate_get("/")

    assert response.json == {"message": "Sentry error."}


@pytest.mark.parametrize("span_streaming", [True, False])
def test_error_in_errorhandler(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[FalconIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    app = falcon.API()

    class Resource:
        def on_get(self, req, resp):
            raise ValueError()

    app.add_route("/", Resource())

    def http500_handler(ex, req, resp, params):
        1 / 0

    app.add_error_handler(Exception, http500_handler)
    client = falcon.testing.TestClient(app)

    if span_streaming:
        items = capture_items("event")

        with pytest.raises(ZeroDivisionError):
            client.simulate_get("/")

        (event,) = (item.payload for item in items)
    else:
        events = capture_events()

        with pytest.raises(ZeroDivisionError):
            client.simulate_get("/")

        (event,) = events

    last_ex_values = event["exception"]["values"][-1]
    assert last_ex_values["type"] == "ZeroDivisionError"
    assert last_ex_values["stacktrace"]["frames"][-1]["vars"]["ex"] == "ValueError()"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_bad_request_not_captured(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[FalconIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    app = falcon.API()

    class Resource:
        def on_get(self, req, resp):
            raise falcon.HTTPBadRequest()

    app.add_route("/", Resource())

    client = falcon.testing.TestClient(app)

    if span_streaming:
        events = capture_items("event")
    else:
        events = capture_events()

    client.simulate_get("/")

    assert not events


@pytest.mark.parametrize("span_streaming", [True, False])
def test_does_not_leak_scope(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[FalconIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    app = falcon.API()

    class Resource:
        def on_get(self, req, resp):
            sentry_sdk.get_isolation_scope().set_tag("request_data", True)

            def generator():
                for row in range(1000):
                    assert sentry_sdk.get_isolation_scope()._tags["request_data"]

                    yield (str(row) + "\n").encode()

            resp.stream = generator()

    app.add_route("/", Resource())

    client = falcon.testing.TestClient(app)

    if span_streaming:
        events = capture_items("event")
    else:
        events = capture_events()

    sentry_sdk.get_isolation_scope().set_tag("request_data", False)

    response = client.simulate_get("/")

    expected_response = "".join(str(row) + "\n" for row in range(1000))
    assert response.text == expected_response
    assert not events
    assert not sentry_sdk.get_isolation_scope()._tags["request_data"]


@pytest.mark.skipif(
    not hasattr(falcon, "asgi"), reason="This Falcon version lacks ASGI support."
)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_falcon_not_breaking_asgi(sentry_init, span_streaming):
    """
    This test simply verifies that the Falcon integration does not break ASGI
    Falcon apps.

    The test does not verify ASGI Falcon support, since our Falcon integration
    currently lacks support for ASGI Falcon apps.
    """
    sentry_init(
        integrations=[FalconIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    asgi_app = falcon.asgi.App()

    try:
        falcon.inspect.inspect_app(asgi_app)
    except TypeError:
        pytest.fail("Falcon integration causing errors in ASGI apps.")


@pytest.mark.skipif(
    (FALCON_VERSION or ()) < (3,),
    reason="The Sentry Falcon integration only supports custom error handlers on Falcon 3+",
)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_falcon_custom_error_handler(
    sentry_init,
    make_app,
    capture_events,
    capture_items,
    span_streaming,
):
    """
    When a custom error handler handles what otherwise would have resulted in a 5xx error,
    changing the HTTP status to a non-5xx status, no error event should be sent to Sentry.
    """
    sentry_init(
        integrations=[FalconIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    app = make_app()
    client = falcon.testing.TestClient(app)

    if span_streaming:
        events = capture_items("event")
    else:
        events = capture_events()

    client.simulate_get("/custom-error")

    assert len(events) == 0


@pytest.mark.parametrize("span_streaming", [True, False])
def test_span_origin(
    sentry_init,
    capture_events,
    capture_items,
    make_client,
    span_streaming,
):
    sentry_init(
        integrations=[FalconIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    client = make_client()

    if span_streaming:
        items = capture_items("span")

        client.simulate_get("/message")

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        assert spans[0]["attributes"]["sentry.origin"] == "auto.http.falcon"
    else:
        events = capture_events()

        client.simulate_get("/message")

        (_, event) = events

        assert event["contexts"]["trace"]["origin"] == "auto.http.falcon"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_falcon_request_media(sentry_init, span_streaming):
    # test_passed stores whether the test has passed.
    test_passed = False

    # test_failure_reason stores the reason why the test failed
    # if test_passed is False. The value is meaningless when
    # test_passed is True.
    test_failure_reason = "test endpoint did not get called"

    class SentryCaptureMiddleware:
        def process_request(self, _req, _resp):
            # This capture message forces Falcon event processors to run
            # before the request handler runs
            sentry_sdk.capture_message("Processing request")

    class RequestMediaResource:
        def on_post(self, req, _):
            nonlocal test_passed, test_failure_reason
            raw_data = req.bounded_stream.read()

            # If the raw_data is empty, the request body stream
            # has been exhausted by the SDK. Test should fail in
            # this case.
            test_passed = raw_data != b""
            test_failure_reason = "request body has been read"

    sentry_init(
        integrations=[FalconIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    try:
        app_class = falcon.App  # Falcon ≥3.0
    except AttributeError:
        app_class = falcon.API  # Falcon <3.0

    app = app_class(middleware=[SentryCaptureMiddleware()])
    app.add_route("/read_body", RequestMediaResource())

    client = falcon.testing.TestClient(app)

    client.simulate_post("/read_body", json={"foo": "bar"})

    # Check that simulate_post actually calls the resource, and
    # that the SDK does not exhaust the request body stream.
    assert test_passed, test_failure_reason



import os
import sys
from unittest import mock

import pytest

import sentry_sdk
from sentry_sdk import get_client
from sentry_sdk.consts import SPANDATA, VERSION


def test_metrics_disabled(sentry_init, capture_envelopes):
    sentry_init(enable_metrics=False)

    envelopes = capture_envelopes()

    sentry_sdk.metrics.count("test.counter", 1)
    sentry_sdk.metrics.gauge("test.gauge", 42)
    sentry_sdk.metrics.distribution("test.distribution", 200)

    assert len(envelopes) == 0


def test_metrics_basics(sentry_init, capture_items):
    sentry_init()
    items = capture_items("trace_metric")

    sentry_sdk.metrics.count("test.counter", 1)
    sentry_sdk.metrics.gauge("test.gauge", 42, unit="millisecond")
    sentry_sdk.metrics.distribution("test.distribution", 200, unit="second")

    get_client().flush()
    metrics = [item.payload for item in items]

    assert len(metrics) == 3

    assert metrics[0]["name"] == "test.counter"
    assert metrics[0]["type"] == "counter"
    assert metrics[0]["value"] == 1.0
    assert "unit" not in metrics[0]
    assert "sentry.sdk.name" in metrics[0]["attributes"]
    assert "sentry.sdk.version" in metrics[0]["attributes"]

    assert metrics[1]["name"] == "test.gauge"
    assert metrics[1]["type"] == "gauge"
    assert metrics[1]["value"] == 42.0
    assert metrics[1]["unit"] == "millisecond"

    assert metrics[2]["name"] == "test.distribution"
    assert metrics[2]["type"] == "distribution"
    assert metrics[2]["value"] == 200.0
    assert metrics[2]["unit"] == "second"


def test_metrics_experimental_option(sentry_init, capture_items):
    sentry_init()
    items = capture_items("trace_metric")

    sentry_sdk.metrics.count("test.counter", 5)

    get_client().flush()

    metrics = [item.payload for item in items]
    assert len(metrics) == 1

    assert metrics[0]["name"] == "test.counter"
    assert metrics[0]["type"] == "counter"
    assert metrics[0]["value"] == 5.0


def test_metrics_with_attributes(sentry_init, capture_items):
    sentry_init(release="1.0.0", environment="test", server_name="test-server")
    items = capture_items("trace_metric")

    sentry_sdk.metrics.count(
        "test.counter", 1, attributes={"endpoint": "/api/test", "status": "success"}
    )

    get_client().flush()

    metrics = [item.payload for item in items]
    assert len(metrics) == 1

    assert metrics[0]["attributes"]["endpoint"] == "/api/test"
    assert metrics[0]["attributes"]["status"] == "success"
    assert metrics[0]["attributes"]["sentry.release"] == "1.0.0"
    assert metrics[0]["attributes"]["sentry.environment"] == "test"

    assert metrics[0]["attributes"][SPANDATA.SERVER_ADDRESS] == "test-server"
    assert metrics[0]["attributes"]["sentry.sdk.name"].startswith("sentry.python")
    assert metrics[0]["attributes"]["sentry.sdk.version"] == VERSION


def test_metrics_with_user(sentry_init, capture_items):
    sentry_init(send_default_pii=True)
    items = capture_items("trace_metric")

    sentry_sdk.set_user(
        {"id": "user-123", "email": "test@example.com", "username": "testuser"}
    )
    sentry_sdk.metrics.count("test.user.counter", 1)

    get_client().flush()

    metrics = [item.payload for item in items]
    assert len(metrics) == 1

    assert metrics[0]["attributes"]["user.id"] == "user-123"
    assert metrics[0]["attributes"]["user.email"] == "test@example.com"
    assert metrics[0]["attributes"]["user.name"] == "testuser"


def test_metrics_no_user_if_pii_off(sentry_init, capture_items):
    sentry_init(send_default_pii=False)
    items = capture_items("trace_metric")

    sentry_sdk.set_user(
        {"id": "user-123", "email": "test@example.com", "username": "testuser"}
    )
    sentry_sdk.metrics.count("test.user.counter", 1)

    get_client().flush()

    metrics = [item.payload for item in items]
    assert len(metrics) == 1

    assert "user.id" not in metrics[0]["attributes"]
    assert "user.email" not in metrics[0]["attributes"]
    assert "user.name" not in metrics[0]["attributes"]


def test_metrics_with_span(sentry_init, capture_items):
    sentry_init(traces_sample_rate=1.0)
    items = capture_items("trace_metric")

    with sentry_sdk.start_transaction(op="test", name="test-span") as transaction:
        sentry_sdk.metrics.count("test.span.counter", 1)

    get_client().flush()

    metrics = [item.payload for item in items]
    assert len(metrics) == 1

    assert metrics[0]["trace_id"] is not None
    assert metrics[0]["trace_id"] == transaction.trace_id
    assert metrics[0]["span_id"] == transaction.span_id


def test_metrics_tracing_without_performance(sentry_init, capture_items):
    sentry_init()
    items = capture_items("trace_metric")

    with sentry_sdk.isolation_scope() as isolation_scope:
        sentry_sdk.metrics.count("test.span.counter", 1)

    get_client().flush()

    metrics = [item.payload for item in items]
    assert len(metrics) == 1

    propagation_context = isolation_scope._propagation_context
    assert propagation_context is not None
    assert metrics[0]["trace_id"] == propagation_context.trace_id
    # Per the metrics spec, span_id is only attached when a span is
    # active when the metric is emitted. The propagation context's
    # synthesized span_id must not be used as a fallback.
    assert "span_id" not in metrics[0]


def test_metrics_before_send(sentry_init, capture_items):
    before_metric_called = False

    def _before_metric(record, hint):
        nonlocal before_metric_called

        assert set(record.keys()) == {
            "timestamp",
            "trace_id",
            "span_id",
            "name",
            "type",
            "value",
            "unit",
            "attributes",
        }

        if record["name"] == "test.skip":
            return None

        before_metric_called = True
        return record

    sentry_init(
        before_send_metric=_before_metric,
    )
    items = capture_items("trace_metric")

    sentry_sdk.metrics.count("test.skip", 1)
    sentry_sdk.metrics.count("test.keep", 1)

    get_client().flush()

    metrics = [item.payload for item in items]
    assert len(metrics) == 1
    assert metrics[0]["name"] == "test.keep"
    assert before_metric_called


def test_metrics_experimental_before_send(sentry_init, capture_items):
    before_metric_called = False

    def _before_metric(record, hint):
        nonlocal before_metric_called

        assert set(record.keys()) == {
            "timestamp",
            "trace_id",
            "span_id",
            "name",
            "type",
            "value",
            "unit",
            "attributes",
        }

        if record["name"] == "test.skip":
            return None

        before_metric_called = True
        return record

    sentry_init(
        _experiments={
            "before_send_metric": _before_metric,
        },
    )
    items = capture_items("trace_metric")

    sentry_sdk.metrics.count("test.skip", 1)
    sentry_sdk.metrics.count("test.keep", 1)

    get_client().flush()

    metrics = [item.payload for item in items]
    assert len(metrics) == 1
    assert metrics[0]["name"] == "test.keep"
    assert before_metric_called


@pytest.mark.tests_internal_exceptions
def test_metrics_before_send_raises_does_not_crash_application(
    sentry_init, capture_items
):
    def _before_metric(record, hint):
        raise ValueError("before_send_metric error")

    sentry_init(
        before_send_metric=_before_metric,
    )
    items = capture_items("trace_metric")

    sentry_sdk.metrics.count("test.keep", 1)

    get_client().flush()

    # The exception in before_send_metric is swallowed and the metric is
    # dropped.
    metrics = [item.payload for item in items]
    assert not metrics


def test_transport_format(sentry_init, capture_envelopes):
    sentry_init(server_name="test-server", release="1.0.0")

    envelopes = capture_envelopes()

    sentry_sdk.metrics.count("test.counter", 1)

    sentry_sdk.get_client().flush()

    assert len(envelopes) == 1
    assert len(envelopes[0].items) == 1
    item = envelopes[0].items[0]

    assert item.type == "trace_metric"
    assert item.headers == {
        "type": "trace_metric",
        "item_count": 1,
        "content_type": "application/vnd.sentry.items.trace-metric+json",
    }
    assert item.payload.json == {
        "version": 2,
        "items": [
            {
                "name": "test.counter",
                "type": "counter",
                "value": 1,
                "timestamp": mock.ANY,
                "trace_id": mock.ANY,
                "attributes": {
                    "sentry.environment": {
                        "type": "string",
                        "value": "production",
                    },
                    "sentry.release": {
                        "type": "string",
                        "value": "1.0.0",
                    },
                    "process.runtime.name": {
                        "type": "string",
                        "value": mock.ANY,
                    },
                    "process.runtime.version": {
                        "type": "string",
                        "value": mock.ANY,
                    },
                    "sentry.sdk.name": {
                        "type": "string",
                        "value": mock.ANY,
                    },
                    "sentry.sdk.version": {
                        "type": "string",
                        "value": VERSION,
                    },
                    "server.address": {
                        "type": "string",
                        "value": "test-server",
                    },
                },
            }
        ],
    }


def test_batcher_drops_metrics(sentry_init, monkeypatch):
    sentry_init()
    client = sentry_sdk.get_client()

    def no_op_flush():
        pass

    monkeypatch.setattr(client.metrics_batcher, "_flush", no_op_flush)

    lost_event_calls = []

    def record_lost_event(reason, data_category, quantity):
        lost_event_calls.append((reason, data_category, quantity))

    monkeypatch.setattr(client.metrics_batcher, "_record_lost_func", record_lost_event)

    for i in range(10_005):  # 5 metrics over the hard limit
        sentry_sdk.metrics.count("test.counter", 1)

    assert len(lost_event_calls) == 5
    for lost_event_call in lost_event_calls:
        assert lost_event_call == ("queue_overflow", "trace_metric", 1)


def test_metric_gets_attributes_from_scopes(sentry_init, capture_items):
    sentry_init()

    items = capture_items("trace_metric")

    global_scope = sentry_sdk.get_global_scope()
    global_scope.set_attribute("global.attribute", "value")

    with sentry_sdk.new_scope() as scope:
        scope.set_attribute("current.attribute", "value")
        sentry_sdk.metrics.count("test", 1)

    sentry_sdk.metrics.count("test", 1)

    get_client().flush()

    metrics = [item.payload for item in items]
    (metric1, metric2) = metrics

    assert metric1["attributes"]["global.attribute"] == "value"
    assert metric1["attributes"]["current.attribute"] == "value"

    assert metric2["attributes"]["global.attribute"] == "value"
    assert "current.attribute" not in metric2["attributes"]


def test_metric_attributes_override_scope_attributes(sentry_init, capture_items):
    sentry_init()

    items = capture_items("trace_metric")

    with sentry_sdk.new_scope() as scope:
        scope.set_attribute("durable.attribute", "value1")
        scope.set_attribute("temp.attribute", "value1")
        sentry_sdk.metrics.count("test", 1, attributes={"temp.attribute": "value2"})

    get_client().flush()

    metrics = [item.payload for item in items]
    (metric,) = metrics

    assert metric["attributes"]["durable.attribute"] == "value1"
    assert metric["attributes"]["temp.attribute"] == "value2"


def test_log_array_attributes(sentry_init, capture_envelopes):
    """Test homogeneous list and tuple attributes, and fallback for inhomogeneous collections."""

    sentry_init()

    envelopes = capture_envelopes()

    with sentry_sdk.new_scope() as scope:
        scope.set_attribute("string_list.attribute", ["value1", "value2"])
        scope.set_attribute("int_tuple.attribute", (3, 2, 1, 4))
        scope.set_attribute("inhomogeneous_tuple.attribute", (3, 2.0, 1, 4))  # type: ignore[arg-type]

        sentry_sdk.metrics.count(
            "test",
            1,
            attributes={
                "float_list.attribute": [3.0, 3.5, 4.2],
                "bool_tuple.attribute": (False, False, True),
                "inhomogeneous_list.attribute": [3.2, True, None],
            },
        )

    get_client().flush()

    assert len(envelopes) == 1
    assert len(envelopes[0].items) == 1
    item = envelopes[0].items[0]
    serialized_attributes = item.payload.json["items"][0]["attributes"]

    assert serialized_attributes["string_list.attribute"] == {
        "value": ["value1", "value2"],
        "type": "array",
    }
    assert serialized_attributes["int_tuple.attribute"] == {
        "value": [3, 2, 1, 4],
        "type": "array",
    }
    assert serialized_attributes["inhomogeneous_tuple.attribute"] == {
        "value": "(3, 2.0, 1, 4)",
        "type": "string",
    }

    assert serialized_attributes["float_list.attribute"] == {
        "value": [3.0, 3.5, 4.2],
        "type": "array",
    }
    assert serialized_attributes["bool_tuple.attribute"] == {
        "value": [False, False, True],
        "type": "array",
    }
    assert serialized_attributes["inhomogeneous_list.attribute"] == {
        "value": "[3.2, True, None]",
        "type": "string",
    }


def test_attributes_preserialized_in_before_send(sentry_init, capture_items):
    """We don't surface user-held references to objects in attributes."""

    def before_send_metric(metric, _):
        assert isinstance(metric["attributes"]["instance"], str)
        assert isinstance(metric["attributes"]["dictionary"], str)
        assert isinstance(metric["attributes"]["inhomogeneous_list"], str)
        assert isinstance(metric["attributes"]["inhomogeneous_tuple"], str)

        return metric

    sentry_init(before_send_metric=before_send_metric)

    items = capture_items("trace_metric")

    class Cat:
        pass

    instance = Cat()
    dictionary = {"color": "tortoiseshell"}

    sentry_sdk.metrics.count(
        "test.counter",
        1,
        attributes={
            "instance": instance,
            "dictionary": dictionary,
            "inhomogeneous_list": [3.2, True, None],
            "inhomogeneous_tuple": (3, 2.0, 1, 4),
        },
    )

    get_client().flush()

    metrics = [item.payload for item in items]
    (metric,) = metrics

    assert isinstance(metric["attributes"]["instance"], str)
    assert isinstance(metric["attributes"]["dictionary"], str)


def test_array_attributes_deep_copied_in_before_send(sentry_init, capture_envelopes):
    """We don't surface user-held references to objects in attributes."""

    strings = ["value1", "value2"]
    ints = (3, 2, 1, 4)

    def before_send_metric(metric, _):
        assert metric["attributes"]["string_list"] is not strings
        assert metric["attributes"]["int_tuple"] is not ints

        return metric

    sentry_init(before_send_metric=before_send_metric)

    sentry_sdk.metrics.count(
        "test.counter",
        1,
        attributes={
            "string_list": strings,
            "int_tuple": ints,
        },
    )

    get_client().flush()


@pytest.mark.skipif(
    sys.platform == "win32"
    or not hasattr(os, "fork")
    or not hasattr(os, "register_at_fork"),
    reason="requires POSIX fork and os.register_at_fork (Python 3.7+)",
)
def test_metrics_batcher_lock_reset_in_child_after_fork(sentry_init):
    """Regression test for the MetricsBatcher fork-deadlock fix.

    If os.fork() runs while another thread holds MetricsBatcher._lock,
    the child inherits the lock locked. The holding thread does not
    exist in the child, so the lock can never be released and
    _ensure_thread deadlocks forever. The after-fork hook must replace
    the lock with a fresh one in the child and reset
    _flusher / _flusher_pid / _buffer / _active / _flush_event.
    """
    sentry_init()
    batcher = sentry_sdk.get_client().metrics_batcher
    assert batcher is not None

    original_lock = batcher._lock
    original_lock.acquire()

    batcher._buffer.append(object())
    batcher._active.flag = True
    batcher._flush_event.set()
    batcher._running = False

    pid = os.fork()
    if pid == 0:
        replaced = batcher._lock is not original_lock
        unheld = batcher._lock.acquire(blocking=False)

        flusher_reset = batcher._flusher is None and batcher._flusher_pid is None
        buffer_reset = len(batcher._buffer) == 0
        active_reset = not getattr(batcher._active, "flag", False)
        event_reset = not batcher._flush_event.is_set()
        running_reset = batcher._running is True

        os._exit(
            0
            if replaced
            and unheld
            and flusher_reset
            and buffer_reset
            and active_reset
            and event_reset
            and running_reset
            else 1
        )

    original_lock.release()
    _, status = os.waitpid(pid, 0)
    assert os.WIFEXITED(status) and os.WEXITSTATUS(status) == 0



"""
Copied from raven-python.

Despite being called "legacy" in some places this resolver is very much still
in use.
"""

import re
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from re import Pattern
    from typing import Dict, List, Optional, Tuple, Union

    from django.urls.resolvers import URLPattern, URLResolver

from django import VERSION as DJANGO_VERSION

if DJANGO_VERSION >= (2, 0):
    from django.urls.resolvers import RoutePattern
else:
    RoutePattern = None

try:
    from django.urls import get_resolver
except ImportError:
    from django.core.urlresolvers import get_resolver


def get_regex(resolver_or_pattern: "Union[URLPattern, URLResolver]") -> "Pattern[str]":
    """Utility method for django's deprecated resolver.regex"""
    try:
        regex = resolver_or_pattern.regex
    except AttributeError:
        regex = resolver_or_pattern.pattern.regex
    return regex


class RavenResolver:
    _new_style_group_matcher = re.compile(
        r"<(?:([^>:]+):)?([^>]+)>"
    )  # https://github.com/django/django/blob/21382e2743d06efbf5623e7c9b6dccf2a325669b/django/urls/resolvers.py#L245-L247
    _optional_group_matcher = re.compile(r"\(\?\:([^\)]+)\)")
    _named_group_matcher = re.compile(r"\(\?P<(\w+)>[^\)]+\)+")
    _non_named_group_matcher = re.compile(r"\([^\)]+\)")
    # [foo|bar|baz]
    _either_option_matcher = re.compile(r"\[([^\]]+)\|([^\]]+)\]")
    _camel_re = re.compile(r"([A-Z]+)([a-z])")

    _cache: "Dict[URLPattern, str]" = {}

    def _simplify(self, pattern: "Union[URLPattern, URLResolver]") -> str:
        r"""
        Clean up urlpattern regexes into something readable by humans:

        From:
        > "^(?P<sport_slug>\w+)/athletes/(?P<athlete_slug>\w+)/$"

        To:
        > "{sport_slug}/athletes/{athlete_slug}/"
        """
        # "new-style" path patterns can be parsed directly without turning them
        # into regexes first
        if (
            RoutePattern is not None
            and hasattr(pattern, "pattern")
            and isinstance(pattern.pattern, RoutePattern)
        ):
            return self._new_style_group_matcher.sub(
                lambda m: "{%s}" % m.group(2), str(pattern.pattern._route)
            )

        result = get_regex(pattern).pattern

        # remove optional params
        # TODO(dcramer): it'd be nice to change these into [%s] but it currently
        # conflicts with the other rules because we're doing regexp matches
        # rather than parsing tokens
        result = self._optional_group_matcher.sub(lambda m: "%s" % m.group(1), result)

        # handle named groups first
        result = self._named_group_matcher.sub(lambda m: "{%s}" % m.group(1), result)

        # handle non-named groups
        result = self._non_named_group_matcher.sub("{var}", result)

        # handle optional params
        result = self._either_option_matcher.sub(lambda m: m.group(1), result)

        # clean up any outstanding regex-y characters.
        result = (
            result.replace("^", "")
            .replace("$", "")
            .replace("?", "")
            .replace("\\A", "")
            .replace("\\Z", "")
            .replace("//", "/")
            .replace("\\", "")
        )

        return result

    def _resolve(
        self,
        resolver: "URLResolver",
        path: str,
        parents: "Optional[List[URLResolver]]" = None,
    ) -> "Optional[str]":
        match = get_regex(resolver).search(path)  # Django < 2.0

        if not match:
            return None

        if parents is None:
            parents = [resolver]
        elif resolver not in parents:
            parents = parents + [resolver]

        new_path = path[match.end() :]
        for pattern in resolver.url_patterns:
            # this is an include()
            if not pattern.callback:
                match_ = self._resolve(pattern, new_path, parents)
                if match_:
                    return match_
                continue
            elif not get_regex(pattern).search(new_path):
                continue

            try:
                return self._cache[pattern]
            except KeyError:
                pass

            prefix = "".join(self._simplify(p) for p in parents)
            result = prefix + self._simplify(pattern)
            if not result.startswith("/"):
                result = "/" + result
            self._cache[pattern] = result
            return result

        return None

    def resolve(
        self,
        path: str,
        urlconf: "Union[None, Tuple[URLPattern, URLPattern, URLResolver], Tuple[URLPattern]]" = None,
    ) -> "Optional[str]":
        resolver = get_resolver(urlconf)
        match = self._resolve(resolver, path)
        return match


LEGACY_RESOLVER = RavenResolver()



from unittest import mock
from unittest.mock import Mock

import pytest

from sentry_sdk.tracing_utils import PropagationContext

SAMPLED_FLAG = {
    None: "",
    False: "-0",
    True: "-1",
}
"""Maps the `sampled` value to the flag appended to the sentry-trace header."""


def test_empty_context():
    ctx = PropagationContext()

    assert ctx.trace_id is not None
    assert len(ctx.trace_id) == 32

    assert ctx.span_id is not None
    assert len(ctx.span_id) == 16

    assert ctx.parent_span_id is None
    assert ctx.parent_sampled is None
    assert ctx.dynamic_sampling_context == {}


def test_context_with_values():
    ctx = PropagationContext(
        trace_id="1234567890abcdef1234567890abcdef",
        span_id="1234567890abcdef",
        parent_span_id="abcdef1234567890",
        parent_sampled=True,
        dynamic_sampling_context={
            "foo": "bar",
        },
    )

    assert ctx.trace_id == "1234567890abcdef1234567890abcdef"
    assert ctx.span_id == "1234567890abcdef"
    assert ctx.parent_span_id == "abcdef1234567890"
    assert ctx.parent_sampled
    assert ctx.dynamic_sampling_context == {
        "foo": "bar",
    }


def test_lazy_uuids():
    ctx = PropagationContext()
    assert ctx._trace_id is None
    assert ctx._span_id is None

    assert ctx.trace_id is not None  # this sets _trace_id
    assert ctx._trace_id is not None
    assert ctx._span_id is None

    assert ctx.span_id is not None  # this sets _span_id
    assert ctx._trace_id is not None
    assert ctx._span_id is not None


def test_property_setters():
    ctx = PropagationContext()

    ctx.trace_id = "X234567890abcdef1234567890abcdef"
    ctx.span_id = "X234567890abcdef"

    assert ctx._trace_id == "X234567890abcdef1234567890abcdef"
    assert ctx.trace_id == "X234567890abcdef1234567890abcdef"
    assert ctx._span_id == "X234567890abcdef"
    assert ctx.span_id == "X234567890abcdef"
    assert ctx.dynamic_sampling_context == {}


def test_update():
    ctx = PropagationContext()

    other_data = {
        "trace_id": "Z234567890abcdef1234567890abcdef",
        "parent_span_id": "Z234567890abcdef",
        "parent_sampled": False,
        "foo": "bar",
    }
    ctx.update(other_data)

    assert ctx._trace_id == "Z234567890abcdef1234567890abcdef"
    assert ctx.trace_id == "Z234567890abcdef1234567890abcdef"
    assert ctx._span_id is None  # this will be set lazily
    assert ctx.span_id is not None  # this sets _span_id
    assert ctx._span_id is not None
    assert ctx.parent_span_id == "Z234567890abcdef"
    assert not ctx.parent_sampled
    assert ctx.dynamic_sampling_context == {}

    assert not hasattr(ctx, "foo")


def test_existing_sample_rand_kept():
    ctx = PropagationContext(
        trace_id="00000000000000000000000000000000",
        dynamic_sampling_context={"sample_rand": "0.5"},
    )

    # If sample_rand was regenerated, the value would be 0.919221 based on the trace_id
    assert ctx.dynamic_sampling_context["sample_rand"] == "0.5"


@pytest.mark.parametrize(
    ("parent_sampled", "sample_rate", "expected_interval"),
    (
        # Note that parent_sampled and sample_rate do not scale the
        # sample_rand value, only determine the range of the value.
        # Expected values are determined by parent_sampled, sample_rate,
        # and the trace_id.
        (None, None, (0.0, 1.0)),
        (None, "0.5", (0.0, 1.0)),
        (False, None, (0.0, 1.0)),
        (True, None, (0.0, 1.0)),
        (False, "0.0", (0.0, 1.0)),
        (False, "0.01", (0.01, 1.0)),
        (True, "0.01", (0.0, 0.01)),
        (False, "0.1", (0.1, 1.0)),
        (True, "0.1", (0.0, 0.1)),
        (False, "0.5", (0.5, 1.0)),
        (True, "0.5", (0.0, 0.5)),
        (True, "1.0", (0.0, 1.0)),
    ),
)
def test_sample_rand_filled(parent_sampled, sample_rate, expected_interval):
    """When continuing a trace, we want to fill in the sample_rand value if it's missing."""
    if sample_rate is not None:
        sample_rate_str = f",sentry-sample_rate={sample_rate}"  # noqa: E231
    else:
        sample_rate_str = ""

    # for convenience, we'll just return the lower bound of the interval as an integer
    mock_randrange = mock.Mock(return_value=int(expected_interval[0] * 1000000))

    def mock_random_class(seed):
        assert seed == "00000000000000000000000000000000", "seed should be the trace_id"
        rv = Mock()
        rv.randrange = mock_randrange
        return rv

    with mock.patch("sentry_sdk.tracing_utils.Random", mock_random_class):
        ctx = PropagationContext().from_incoming_data(
            {
                "sentry-trace": f"00000000000000000000000000000000-0000000000000000{SAMPLED_FLAG[parent_sampled]}",
                # Placeholder is needed, since we only add sample_rand if sentry items are present in baggage
                "baggage": f"sentry-placeholder=asdf{sample_rate_str}",
            }
        )

    assert (
        ctx.dynamic_sampling_context["sample_rand"] == f"{expected_interval[0]:.6f}"  # noqa: E231
    )
    assert mock_randrange.call_count == 1
    assert mock_randrange.call_args[0] == (
        int(expected_interval[0] * 1000000),
        int(expected_interval[1] * 1000000),
    )


def test_sample_rand_rounds_down():
    # Mock value that should round down to 0.999_999
    mock_randrange = mock.Mock(return_value=999999)

    def mock_random_class(_):
        rv = Mock()
        rv.randrange = mock_randrange
        return rv

    with mock.patch("sentry_sdk.tracing_utils.Random", mock_random_class):
        ctx = PropagationContext().from_incoming_data(
            {
                "sentry-trace": "00000000000000000000000000000000-0000000000000000",
                "baggage": "sentry-placeholder=asdf",
            }
        )

    assert ctx.dynamic_sampling_context["sample_rand"] == "0.999999"
